#include <sys/types.h>
#include <sys/time.h>   // POSIX place for select
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <ctype.h>
#include <errno.h>
#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <assert.h>

#ifdef NO_NCURSES
#include <curses.h>
#else
#include <ncurses/curses.h>
#endif

#ifdef DECLARE_SNPRINTF
extern "C" {
    int snprintf(char *, size_t, const char *, ...);
    int vsnprintf(char *, size_t, const char *, va_list);
}
#endif

#include "primarray.h"
#include "ptrarray.h"
#include "version.h"

#define BOARDSIZE  15
#define RACKSIZE   7
#define OLD_PROTO_VER  1
#define PROTO_VER  2
#define SCRAB_PORT 7096

static int old_protocol=0;
static int input8bit=0;

////////////////////////////////////////////////////////////

static char ourserver[32];
static int sock;

static void sendit(const char *fmt, ...) {
    char buf[4096];
    va_list ap;
    va_start(ap, fmt);
    vsnprintf(buf, sizeof(buf), fmt, ap);
    va_end(ap);
    write(sock, buf, strlen(buf));
}

////////////////////////////////////////////////////////////

typedef enum { B_NONE, B_2L, B_3L, B_2W, B_3W } bonustypes;

struct boardentry {
    bonustypes bonus;      /* what's on this square */
    int bold;              /* if boldfacing this space */
    int tile;              /* A-Z or _ or . for empty */
    int phonytile;         /* a tile put down only in the editor, or '.' */
};

static boardentry board[BOARDSIZE][BOARDSIZE];
static int boardxpos=BOARDSIZE/2, boardypos=BOARDSIZE/2;
static int hidebonuses=0;

static void initboard() {
    for (int x=0; x<BOARDSIZE; x++) for (int y=0; y<BOARDSIZE; y++) {
	board[x][y].bonus = B_NONE;
	board[x][y].phonytile = '.';
	board[x][y].bold = 0;
	board[x][y].tile = '.';
    }
    for (int x=0; x<BOARDSIZE; x+=BOARDSIZE/2) {
	// triple word scores
	for (int y=0; y<BOARDSIZE; y+=BOARDSIZE/2) {
	    board[x][y].bonus = B_3W;
	}
    }
    // double word scores
    board[BOARDSIZE/2][BOARDSIZE/2].bonus = B_2W;
    for (int i=1; i<=4; i++) {
	board[i][i].bonus = B_2W;
	board[BOARDSIZE-i-1][i].bonus = B_2W;
	board[i][BOARDSIZE-i-1].bonus = B_2W;
	board[BOARDSIZE-i-1][BOARDSIZE-i-1].bonus = B_2W;
    }
    // triple letter scores
    for (int x=1; x<BOARDSIZE; x+=4) for (int y=1; y<BOARDSIZE; y+=4) {
	if (board[x][y].bonus==B_NONE) board[x][y].bonus = B_3L;
    }
    // double letter scores
#define SETDOUB(x,y) \
    board[x][y].bonus = B_2L; board[y][x].bonus = B_2L; \
    board[BOARDSIZE-x-1][y].bonus=B_2L; board[y][BOARDSIZE-x-1].bonus=B_2L; \
    board[x][BOARDSIZE-y-1].bonus=B_2L; board[BOARDSIZE-y-1][x].bonus=B_2L; \
    board[BOARDSIZE-x-1][BOARDSIZE-y-1].bonus = B_2L; \
    board[BOARDSIZE-y-1][BOARDSIZE-x-1].bonus = B_2L;

    SETDOUB(3,0);
    SETDOUB(6,2);
    SETDOUB(7,3);
    SETDOUB(6,6);

#undef SETDOUB
}

static void clearbold() {
    for (int x=0; x<BOARDSIZE; x++) for (int y=0; y<BOARDSIZE; y++) {
	board[x][y].bold = 0;
    }
}

static void update_board_line(int y, const char *tiles) {
    for (int x=0; tiles[x]; x++) {
	board[x][y].tile = tiles[x];
    }
}

////////////////////////////////////////////////////////////

struct userinfo {
    char name[16];
    int score;
};

static char ourname[16];
static char attempted_name[16];
static char whoseturn[16];
static int ourturn=0;
static int observer=0;
static primarray<userinfo> users;

static void reset_scores() {
    users.setsize(0);
}

static void update_score(const char *name, int score) {
    for (int i=0; i<users.num(); i++) {
	if (!strcasecmp(name, users[i].name)) {
	    users[i].score = score;
	    return;
	}
    }
    userinfo u;
    strncpy(u.name, name, sizeof(u.name));
    u.name[sizeof(u.name)-1] = 0;
    u.score = score;
    users.add(u);
}

static void handle_rename(const char *oldname, const char *newname) {
    for (int i=0; i<users.num(); i++) {
	if (!strcasecmp(oldname, users[i].name)) {
	    strncpy(users[i].name, newname, sizeof(users[i].name));
	    users[i].name[sizeof(users[i].name)-1] = 0;
	}
    }
}

////////////////////////////////////////////////////////////

static char ourrack[RACKSIZE+1];
static int rackpos=0;

static void add_rack(int tile) {
    char foo[2];
    foo[0] = tile;
    foo[1] = 0;
    if (strlen(ourrack)>6) return;
    strcat(ourrack, foo);
}

static void set_rack(const char *rack) {
    strncpy(ourrack, rack, sizeof(ourrack));
    ourrack[sizeof(ourrack)-1] = 0;
    int racklen = strlen(ourrack);
    if (rackpos >= racklen) rackpos = racklen;
}

static void swap_rack(int p1, int p2) {
    int racklen = strlen(ourrack);
    assert(p1 >= 0 && p1 < racklen);
    assert(p2 >= 0 && p2 < racklen);

    char tmp = ourrack[p1];
    ourrack[p1] = ourrack[p2];
    ourrack[p2] = tmp;
}

static void randomize_rack() {
   primarray<char> letters;
   for (int i=0; ourrack[i]; i++) {
      letters.add(ourrack[i]);
   }

   ourrack[0] = 0;
   while (letters.num() > 0) {
      int ix = random() % letters.num();
      int tile = letters[ix];
      if (ix < letters.num()-1) {
	 letters[ix] = letters[letters.num()-1];
      }
      letters.setsize(letters.num()-1);
      add_rack(tile);
   }
}

////////////////////////////////////////////////////////////

static ptrarray<char> loglines;
static int logscrollup=0;

static void dolog(const char *fmt, ...) {
    char buf[4096];
    va_list ap;
    va_start(ap, fmt);
    vsnprintf(buf, sizeof(buf), fmt, ap);
    va_end(ap);

    ptrarray<char> sublines;
    int width = COLS-1;
    int space=-1, front=0;
    for (int i=0; buf[i]; i++) {
	if (buf[i]==' ') space=i;
	if (i-front>=width && space>=0) {
	    sublines.add(buf+front);
	    buf[space]=0;
	    front = space+1;
	    space = -1;
	    width = COLS-3;
	}
    }
    sublines.add(buf+front);

    char *x = new char[strlen(sublines[0])+1];
    strcpy(x, sublines[0]);
    loglines.add(x);

    for (int i=1; i<sublines.num(); i++) {
	char *x = new char[strlen(sublines[i])+3];
	strcpy(x, "  ");
	strcat(x, sublines[i]);
	loglines.add(x);
    }
}

////////////////////////////////////////////////////////////

static char inputline[512];
static int inputpos=0;
static enum { I_SAY, I_EMOTE, I_TELL, I_NAME, I_RAW } inputtype = I_SAY;

static void sendmessage() {
    if (inputtype==I_TELL) {
	char *to = inputline;
	while (*to==' ') to++;
	char *txt = strchr(to, ' ');
	if (!txt) return;
	*(txt++) = 0;
	if (strlen(txt)==0) return;
	sendit("TELL %s %s\r\n", to, txt);
	dolog("-> *%s* %s", to, txt);
    }
    else if (inputtype==I_SAY) {
	if (strlen(inputline)==0) return;
	sendit("SAY %s\r\n", inputline);
    }
    else if (inputtype==I_EMOTE) {
	if (strlen(inputline)==0) return;
	sendit("EMOTE %s\r\n", inputline);
    }
    else if (inputtype==I_NAME) {
	if (strlen(inputline)==0) return;
	if (strchr(inputline, ' ') || strlen(inputline) >= sizeof(ourname)) {
	    dolog("*** Illegal name");
	    return;
	}
	strcpy(attempted_name, inputline);
	sendit("NAME %s\r\n", inputline);
    }
    else {
	if (strlen(inputline)==0) return;
	sendit("%s\r\n", inputline);
    }
}

static int back_a_word(int pos) {
   while (pos > 0 && inputline[pos-1]==' ') pos--;
   while (pos > 0 && inputline[pos-1]!=' ') pos--;
   return pos;
}

static int forward_a_word(int pos) {
   while (inputline[pos]==' ') pos++;
   while (inputline[pos]!=' ' && inputline[pos]!=0) pos++;
   return pos;
}

static int inputkey(int ch) {
    if (ch>=32 && ch<127 || (ch>=32+128 && ch<127+128)) {
	if (strlen(inputline) < sizeof(inputline)-1) {
	    memmove(inputline+inputpos+1, inputline+inputpos,
		    strlen(inputline+inputpos)+1);
	    inputline[inputpos++] = ch;
	}
	else beep();
    }
    else if (ch==1) { /* ^A */ inputpos=0; }
    else if (ch==2) { /* ^B */ if (inputpos>0) inputpos--; }
    else if (ch==4) { /* ^D */ 
	if (strlen(inputline+inputpos)>0) {
	    memmove(inputline+inputpos, inputline+inputpos+1,
		    strlen(inputline+inputpos+1)+1);
	}
	else beep();
    }
    else if (ch==5) { /* ^E */ inputpos = strlen(inputline); }
    else if (ch==6) { /* ^F */ if (inputline[inputpos]) inputpos++; }
    else if (ch==7) { /* ^G */
       beep();
       inputline[0] = 0;
       inputpos = 0;
       return 1;
    }
    else if (ch==8 || ch==127) { /* ^H */ 
	if (inputpos>0) {
	    inputpos--;
	    memmove(inputline+inputpos, inputline+inputpos+1,
		    strlen(inputline+inputpos+1)+1);
	}
	else beep();
    }
    else if (ch==10 || ch==13) { /* ^J / ^M */
	sendmessage();
	inputline[0] = 0;
	inputpos = 0;
	return 1;
    }
    else if (ch==11) { /* ^K */ inputline[inputpos]=0; }
    else if (ch==14) { /* ^N */ }
    else if (ch==16) { /* ^P */ }
    else if (ch==20) { /* ^T */ 
	if (inputpos>0 && inputline[inputpos]) {
	    int tmp = inputline[inputpos-1];
	    inputline[inputpos-1] = inputline[inputpos];
	    inputline[inputpos] = tmp;
	    inputpos++;
	}
    }
    else if (ch==21) { /* ^U */
       inputpos = 0;
       inputline[inputpos] = 0;
    }
    else if (ch==23 || ch==8+256 || ch==127+256) { /* ^W or alt-^H */
       int otherpos = back_a_word(inputpos);
       if (otherpos < inputpos) {
	  memmove(inputline+otherpos, inputline+inputpos, 
		  strlen(inputline+inputpos)+1);
       }
       inputpos = otherpos;
    }
    else if (ch=='b'+256) { /* alt-b */ inputpos = back_a_word(inputpos); }
    else if (ch=='d'+256) { /* alt-d */
       int otherpos = forward_a_word(inputpos);
       if (inputpos < otherpos) {
	  memmove(inputline+inputpos, inputline+otherpos, 
		  strlen(inputline+otherpos)+1);
       }
    }
    else if (ch=='f'+256) { /* alt-f */ inputpos = forward_a_word(inputpos); }

    return 0;
}

////////////////////////////////////////////////////////////

struct gameinfo {
    int exists;
    int started;
    int players;
    int observers;
};

static primarray<gameinfo> gamelist;
static primarray<int> livegamelist;  /* indexes into gamelist */
static int gamelistscroll=0;
static int gamelistcursor=0;
static int in_game=0;

static void update_live_games() {
    livegamelist.setsize(0);
    for (int i=0; i<gamelist.num(); i++) {
	if (gamelist[i].exists) livegamelist.add(i);
    }
    while (gamelistscroll + gamelistcursor > livegamelist.num()) {
	if (gamelistscroll > 0) gamelistscroll--;
	else gamelistcursor--;
    }
}

static void game_exists(int n) {
    if (n>=gamelist.num()) {
	int q = gamelist.num();
	gamelist.setsize(n+1);
	for (int i=q; i<=n; i++) {
	    gamelist[i].exists = 0;
	    gamelist[i].started = 0;
	    gamelist[i].players = 0;
	    gamelist[i].observers = 0;
	}
    }
    gamelist[n].exists = 1;
}

static void gamelistkey(int ch) {
    if (ch==10 || ch==13) { /* ^J/^M */
	int ix = gamelistcursor+gamelistscroll;
	assert(ix>=0 && ix<=livegamelist.num());
	if (ix==livegamelist.num()) {
	    sendit("NEWGAME\r\n");
	}
	else {
	    int gamenum = livegamelist[ix];
	    sendit("JOIN %d\r\n", gamenum);
	}
    }
    else if (ch==14) { /* ^N */
	if (gamelistcursor+gamelistscroll >= livegamelist.num()) return;
	if (gamelistcursor < BOARDSIZE-4) gamelistcursor++;
	else gamelistscroll++;
    }
    else if (ch==16) { /* ^P */
	if (gamelistcursor > 0) gamelistcursor--;
	else if (gamelistscroll > 0) gamelistscroll--;
    }
}

////////////////////////////////////////////////////////////

static enum { 
    M_NOTREADY, M_READY, M_BOARD, M_RACK
} mode;
static int inputmode=0;
static int needapprove=0;
static int showcounts=0;

static int canusecolor=0;
static int usecolor=0;
static int colorbold=0;
#define WORD2ATTR   COLOR_PAIR(1)
#define WORD3ATTR   COLOR_PAIR(2)
#define LETR2ATTR   COLOR_PAIR(3)
#define LETR3ATTR   COLOR_PAIR(4)
#define EXTRABOLD   COLOR_PAIR(5)

static void boldon() {
   if (usecolor && colorbold) {
      attron(EXTRABOLD);
   }
   else {
      attron(A_BOLD);
   }
}

static void boldoff() {
   if (usecolor && colorbold) {
      attroff(EXTRABOLD);
   }
   else {
      attroff(A_BOLD);
   }
}

static void redraw() {
    erase();

    if (in_game) {
	/* board */
	for (int x=0; x<BOARDSIZE; x++) for (int y=0; y<BOARDSIZE; y++) {
	    move(y,x*2);
	    boardentry b = board[x][y];
	    if (b.phonytile!='.') { 
		attron(A_REVERSE);
		addch(b.phonytile);
		attroff(A_REVERSE);
	    }
	    else if (b.tile != '.') {
		if (b.bold) boldon();
		addch(b.tile);
		if (b.bold) boldoff();
	    }
	    else if (b.bonus==B_2W && !hidebonuses) {
		if (usecolor) {
		    attron(WORD2ATTR); addch('#'); attroff(WORD2ATTR);
		}
		else {
		    attron(A_DIM); addch('%'); attroff(A_DIM);
		}
	    }
	    else if (b.bonus==B_3W && !hidebonuses) {
		if (usecolor) {
		    attron(WORD3ATTR); addch('#'); attroff(WORD3ATTR);
		}
		else {
		    attron(A_DIM); addch('&'); attroff(A_DIM);
		}
	    }
	    else if (b.bonus==B_2L && !hidebonuses) {
		if (usecolor) {
		    attron(LETR2ATTR); addch('#'); attroff(LETR2ATTR);
		}
		else {
		    attron(A_DIM); addch('2'); attroff(A_DIM);
		}
	    }
	    else if (b.bonus==B_3L && !hidebonuses) {
		if (usecolor) {
		    attron(LETR3ATTR); addch('#'); attroff(LETR3ATTR);
		}
		else {
		    attron(A_DIM); addch('3'); attroff(A_DIM);
		}
	    }
	    else addch(' ');
	}
    }
    else {
	/* game list */
	mvaddstr(0, 10, "Game list");
	mvaddstr(1, 0, "##  State     Players   Obs.");
	for (int y=0; y<BOARDSIZE-3; y++) {
	    int line = y+gamelistscroll;
	    move(y+3, 0);
	    if (y==gamelistcursor) attron(A_REVERSE);

	    char buf[29];
	    const char *txt = "";

	    if (line>=0 && line<livegamelist.num()) {
		int gamenum = livegamelist[line];
		gameinfo *g = &gamelist[gamenum];

		snprintf(buf, sizeof(buf), "%2d  %s   %-2d      %d",
			 gamenum, g->started ? "playing" : "forming",
			 g->players, g->observers);

		txt = buf;
	    }
	    else if (line==livegamelist.num()) {
		txt = "        Form new game";
	    }
	    addstr(txt);
	    for (int k=strlen(txt); k<29; k++) addch(' ');
	    if (y==gamelistcursor) attroff(A_REVERSE);
	    clrtoeol();
	}
    }

    /* scores */
    for (int i=0; i<users.num(); i++) {
	move(i, 40);
	if (!strcasecmp(users[i].name, whoseturn)) boldon();
	addstr(users[i].name);
	if (!strcasecmp(users[i].name, whoseturn)) boldoff();
	char tmp[16];
	snprintf(tmp, sizeof(tmp), "%d ", users[i].score);
	move(i, 60);
	addstr(tmp);
    }

    /* point values */
    if (showcounts) {
	mvaddstr(7, 40,  "       Letter Counts");
	mvaddstr(9, 40,  "BCFHMPVWY: 2  G: 3  DLSU: 4");
	mvaddstr(10, 40, "NRT: 6 O: 8  AI: 9  E: 12");
	mvaddstr(11, 40, "Blank (_): 2       JKQXZ: 1");
    }
    else {
	mvaddstr(7, 40,  "         Scoring");
	mvaddstr(9, 40,  "AEIOULNRST: 1  DG: 2  BCMP: 3");
	mvaddstr(10, 40, "FHVWY: 4  K: 5  JX: 8  QZ: 10");
	mvaddstr(11, 40, "Blank (_): 0");
	mvaddstr(12, 40, "50 point bonus for 7-tile plays");
	mvaddstr(13, 40, " /  are 2x/3x letter multipliers");
	mvaddstr(14, 40, " /  are 2x/3x word multipliers");
	if (usecolor) {
	    attron(LETR2ATTR); mvaddch(13,40,'#'); attroff(LETR2ATTR);
	    attron(LETR3ATTR); mvaddch(13,42,'#'); attroff(LETR3ATTR);
	    attron(WORD2ATTR); mvaddch(14,40,'#'); attroff(WORD2ATTR);
	    attron(WORD3ATTR); mvaddch(14,42,'#'); attroff(WORD3ATTR);
	}
	else {
	    attron(A_DIM);
	    mvaddch(13,40,'2');
	    mvaddch(13,42,'3');
	    mvaddch(14,40,'%');
	    mvaddch(14,42,'&');
	    attroff(A_DIM);
	}
    }
	
    /* log */
    for (int y=BOARDSIZE; y<LINES-2; y++) {
	int line = y-(LINES-2)+loglines.num() - logscrollup;
	move(y, 0);
	if (line>=0 && line < loglines.num()) {
	    addstr(loglines[line]);
	}
	clrtoeol();
    }

    /* name */
    move(LINES-2, 0);
    {
	attron(A_REVERSE);
	char tmp[30];
	if (logscrollup>0) {
	    snprintf(tmp, sizeof(tmp), "  %-30s", "[scrolled up]");
	}
	else {
	    snprintf(tmp, sizeof(tmp), "  %s@%-30s", ourname, ourserver);
	}
	addstr(tmp);
	attroff(A_REVERSE);
    }
    clrtoeol();

    /* rack */
    if (mode==M_BOARD || mode==M_RACK) {
	move(LINES-2, 30);
	addstr("Letters: ");
	addstr(ourrack);
	clrtoeol();
    }

    /* approval */
    if (needapprove) {
	move(LINES-2, COLS-20);
	attron(A_REVERSE);
	addstr("press ^O to approve");
	attroff(A_REVERSE);
	clrtoeol();
    }
    else if (mode==M_NOTREADY && in_game) {
	move(LINES-2, COLS-18);
	attron(A_REVERSE);
	addstr("press ^R to start");
	attroff(A_REVERSE);
	clrtoeol();
    }

    /* input line */
    int vis_inputpos=0;
    move(LINES-1, 0);
    if (inputmode) {
	const char *inputprompt = "";
	if (inputtype==I_SAY) inputprompt = "/say: ";
	else if (inputtype==I_EMOTE) {
	    char emoteprompt[128];
	    snprintf(emoteprompt, sizeof(emoteprompt), ": * %s ", ourname);
	    inputprompt = emoteprompt;
	}
	else if (inputtype==I_TELL) inputprompt = ">tell ";
	else if (inputtype==I_NAME) inputprompt = "<name: ";
	else inputprompt = "\\raw: ";

	int blocksize = COLS-20;
	int trigger = COLS-10;
	size_t offset = 0;
	vis_inputpos = inputpos+strlen(inputprompt);
	while (vis_inputpos > trigger) {
	    vis_inputpos -= blocksize;
	    offset += blocksize;
	}

	if (offset < strlen(inputprompt)) {
	    addstr(inputprompt+offset);
	    addstr(inputline);
	}
	else {
	    offset -= strlen(inputprompt);
	    addstr(inputline + offset);
	}
    }
    clrtoeol();

    /* final cursor position */
    if (inputmode) {
	move(LINES-1, vis_inputpos);
    }
    else if (!in_game) {
	move(gamelistcursor+3, 28);
    }
    else if (mode == M_BOARD) {
	move(boardypos, boardxpos*2);
    }
    else if (mode == M_RACK) {
	move(LINES-2, rackpos+39);
    }
    else {
	move(LINES-1, 0);
    }

    refresh();
}

static void boardkey(int ch) {
    ch = toupper(ch);
    if ((ch>='A' && ch<='Z') || ch=='_') {
	ch = toupper(ch);
	
	if (board[boardxpos][boardypos].tile == '.') {
	    board[boardxpos][boardypos].phonytile = ch;
	    //if (boardxpos < BOARDSIZE-1) boardxpos++;
	}
	else beep();
    }
    else if (ch==' ') {
	if (board[boardxpos][boardypos].phonytile!='.') {
	    board[boardxpos][boardypos].phonytile = '.';
	    //if (boardxpos < BOARDSIZE-1) boardxpos++;
	}
	else beep();
    }
    else if (ch == 1) { /* ^A */ boardxpos = 0; }
    else if (ch == 2) { /* ^B */ if (boardxpos>0) boardxpos--; }
    else if (ch == 5) { /* ^E */ boardxpos = BOARDSIZE-1; }
    else if (ch == 6) { /* ^F */ if (boardxpos<BOARDSIZE-1) boardxpos++; }
    else if (ch == 10 || ch == 13) { /* ^J/^M */ 
	bool first=true;
	for (int x=0; x<BOARDSIZE; x++) for (int y=0; y<BOARDSIZE; y++) {
	    if (board[x][y].phonytile!='.') {
		if (first) { sendit("PLAY"); first = false; }
		sendit(" %d,%d,%c", x, y, board[x][y].phonytile);
	    }
	}
	if (!first) sendit("\r\n");
    }
    else if (ch == 14) { /* ^N */ if (boardypos<BOARDSIZE-1) boardypos++; }
    else if (ch == 16) { /* ^P */ if (boardypos>0) boardypos--; }
    else if (ch == '\t') mode = M_RACK;
}

static void rackkey(int ch) {
    int racklen = strlen(ourrack);  /* note: keep this signed */

    if (ch == '\t') {
       mode = M_BOARD;
    }
    else if (ch == 1) { /* ^A */
       rackpos = 0;
    }
    else if (ch == 2) { /* ^B (left) */
       if (rackpos > 0) {
	  rackpos--;
       }
    }
    else if (ch == 'b'+256) { /* M-b */
       if (rackpos > 0) {
	  if (rackpos < racklen) swap_rack(rackpos, rackpos-1);
	  rackpos--;
       }
    }
    else if (ch == 5) { /* ^E */
       if (racklen < 1) {
	  rackpos = 0;
       }
       else {
	  rackpos = racklen-1;
       }
    }
    else if (ch == 6) { /* ^F (right) */
       if (rackpos < racklen-1) {
	  rackpos++;
       }
    }
    else if (ch == 'f'+256) { /* M-f */
       if (rackpos < racklen-1) {
	  swap_rack(rackpos, rackpos+1);
	  rackpos++;
       }
    }
    else if (ch == 20) { /* ^T */
       if (rackpos > 0 && rackpos < racklen) {
	  swap_rack(rackpos, rackpos-1);
	  if (rackpos < racklen-1) rackpos++;
       }
    }
    else if ((ch >= 'A' && ch <= 'Z') || (ch >= 'a' && ch <= 'z')) {
       ch = toupper(ch);
       int pos = -1;
       for (int i=rackpos; ourrack[i]; i++) {
	  if (ourrack[i] == ch) {
	     pos = i;
	     break;
	  }
       }
       if (pos >= 0) {
	  swap_rack(rackpos, pos);
	  int racklen = strlen(ourrack);  /* note: keep this signed */
	  if (rackpos < racklen-1) {
	     rackpos++;
	  }
       }
       else beep();
    }
}

static void keystroke(int ch) {
    if (ch==12) { /* ^L */ clear(); }
    else if (ch=='p'+256) { /* M-p */ 
	int lines = LINES-2-BOARDSIZE-1;
	logscrollup += lines;
	if (logscrollup>loglines.num()-1) logscrollup = loglines.num()-1;
	if (logscrollup<0) logscrollup=0;
    }
    else if (ch=='n'+256) { /* M-n */ 
	int lines = LINES-2-BOARDSIZE-1;
	logscrollup -= lines;
	if (logscrollup<0) logscrollup=0;
    }
    else if (ch=='<'+256) { /* M-< */ 
	logscrollup = loglines.num()-1;
	if (logscrollup<0) logscrollup=0;
    }
    else if (ch=='>'+256) { /* M-> */ 
	logscrollup = 0;
    }
    else if (ch=='h'+256) { /* M-h */ hidebonuses = !hidebonuses; }
    else if (ch=='c'+256 && canusecolor) { /* M-c */ usecolor = !usecolor; }
    else if (ch=='s'+256) { /* M-s */ showcounts = !showcounts; }
    else if (ch == 'r'+256) { /* M-r */ randomize_rack(); }
    else if (mode==M_NOTREADY && in_game && ch==18) { /* ^R */ 
	sendit("READY\r\n");
    }
    else if (mode==M_NOTREADY && in_game && ch==23) { /* ^W */ 
	sendit("OBSERVER\r\n");
    }
    else if (needapprove && (ch==15 || ch=='o'+256)) { 
	/* ^O or M-O */
	sendit("APPROVE\r\n");
	needapprove=0; 
    }
    else if (!inputmode && ch=='!') { inputmode = 1; inputtype = I_TELL; }
    else if (!inputmode && ch=='.') { inputmode = 1; inputtype = I_TELL; }
    else if (!inputmode && ch=='>') { inputmode = 1; inputtype = I_TELL; }
    else if (!inputmode && ch=='/') { inputmode = 1; inputtype = I_SAY; }
    else if (!inputmode && ch==':') { inputmode = 1; inputtype = I_EMOTE; }
    else if (!inputmode && ch==',') { inputmode = 1; inputtype = I_NAME; }
    else if (!inputmode && ch=='<') { inputmode = 1; inputtype = I_NAME; }
    else if (!inputmode && ch=='\\') { inputmode = 1; inputtype = I_RAW; }
    else if (inputmode) {
	if (inputkey(ch)) inputmode = 0;
    }
    else if (!in_game) gamelistkey(ch);
    else if (mode==M_BOARD) boardkey(ch);
    else if (mode==M_RACK) rackkey(ch);
}

static void raw_keystroke(int ch) {
    static enum { 
	S_NORM, S_GOTESC, S_GOTESCO, S_GOTESCBRACKET
    } state = S_NORM;
    static int counter=0;

    if (!input8bit) {
	/* treat 8th bit as meta bit */
	ch = (ch & 0x7f) | ((ch & 0x80)<<1);
    }

    if (state==S_NORM) {
	if (ch==27) state = S_GOTESC;
	else keystroke(ch);
    }
    else if (state==S_GOTESC) {
	if (ch=='O') state = S_GOTESCO;
	else if (ch=='[') state = S_GOTESCBRACKET;
	else {
	    keystroke(ch|0x100);
	    state = S_NORM;
	}
    }
    else if (state==S_GOTESCO) {
	switch (ch) {
	  case 'A': keystroke(16); break;  /* up -> ^P */
	  case 'B': keystroke(14); break;  /* dn -> ^N */
	  case 'C': keystroke(6); break;  /* right -> ^F */
	  case 'D': keystroke(2); break;  /* left -> ^B */
	  case 'q': keystroke(14); keystroke(2); break;  /* end -> ^N^B */
	  case 'r': keystroke(14); break;  /* dn -> ^N */
	  case 's': keystroke(14); keystroke(6); break;  /* pgdn -> ^N^F */
	  case 't': keystroke(2); break;  /* left -> ^B */
	  case 'v': keystroke(6); break;  /* right -> ^F */
	  case 'w': keystroke(16); keystroke(2); break;  /* home -> ^P^B */
	  case 'x': keystroke(16); break;  /* up -> ^P */
	  case 'y': keystroke(16); keystroke(6); break;  /* pgup -> ^P^F */
	}
	state = S_NORM;
    }
    else if (state==S_GOTESCBRACKET) {
	if (isdigit(ch)) {
	    counter = counter*10+ch-'0';
	    return;
	}
	switch (ch) {
	  case 'A': keystroke(16); break;  /* up -> ^P */
	  case 'B': keystroke(14); break;  /* dn -> ^N */
	  case 'C': keystroke(6); break;  /* right -> ^F */
	  case 'D': keystroke(2); break;  /* left -> ^B */
	  case '~': switch (counter) {
	    case 1: keystroke(16); keystroke(2); break;  /* home -> ^P^B */
	    case 5: keystroke(16); keystroke(6); break;  /* pgup -> ^P^F */
	    case 4: keystroke(14); keystroke(2); break;  /* end -> ^N^B */
	    case 6: keystroke(14); keystroke(6); break;  /* pgdn -> ^N^F */
	  }
	  break;
	}
	state = S_NORM;
	counter = 0;
    }
    redraw();
}

////////////////////////////////////////////////////////////

static void do_logerr(const char *line) {
    const char *x = strchr(line, ' ');
    if (x) {
	while (*x==' ') x++;
	dolog("*** %s", x);
    }
    redraw();
}

static void do_1000(const char *line) {
    const char *x = strchr(line, ' ');
    if (!x) {
	printf("Ill-formed server banner\n");
	exit(1);
    }
    while (*x==' ') x++;
    int ver = atoi(x);
    if (ver==OLD_PROTO_VER) {
	old_protocol=1;
	in_game = 1;
    }
    else if (ver!=PROTO_VER) {
	printf("Protocol version was %d, expected %d\n", ver, PROTO_VER);
	printf("Can't deal with this, sorry\n");
	exit(1);
    }
    x = strchr(x, ' ');

    initscr();
    cbreak();
    noecho();
    if (has_colors()) {
	canusecolor = 1;
	/* initial usecolor set from command line */
    }
    else {
	usecolor = 0;
    }
    if (canusecolor) {
	start_color();
	init_pair(1, COLOR_YELLOW, COLOR_BLACK);    /* double word */
	init_pair(2, COLOR_RED, COLOR_BLACK);  /* triple word */
	init_pair(3, COLOR_CYAN, COLOR_BLACK);   /* double letter */
	init_pair(4, COLOR_BLUE, COLOR_BLACK);   /* triple letter */
	init_pair(5, COLOR_GREEN, COLOR_BLACK);   /* boldface */
    }
    //leaveok(stdscr, FALSE);

    dolog("*** Client version %s", VERSION);

    if (x) {
	while (*x==' ') x++;
	dolog("*** %s", x);
    }

    strcpy(attempted_name, ourname);
    sendit("NAME %s\r\n", ourname);
    if (!old_protocol) sendit("GLIST\r\n");

    redraw();
}

static void do_500x(const char *line) {
    const char *x = strchr(line, ' ');
    if (!x) {
	printf("Ill-formed server banner\n");
	exit(1);
    }
    while (*x==' ') x++;
    printf("Server refused to talk to us, saying %s\n", x);
    exit(1);
}

static void do_1010(const char *) {
    strcpy(ourname, attempted_name);
}

static void do_1020_2001(const char *line) {
    const char *x = strchr(line, ' ');
    if (x) {
	while (*x==' ') x++;
	int tile = *x;
	if (tile=='_') dolog("*** Drawing for play order: got a blank");
	else dolog("*** Drawing for play order: got %c", tile);
	if (mode == M_NOTREADY) mode = M_READY;
    }
    redraw();
}

static void do_1120_2020(const char *line) {
    observer=1;
    if (mode==M_NOTREADY) mode = M_READY;
    dolog("*** Observer mode.");
    if (atoi(line)==2020) sendit("REFRESH\r\n");  // if game already started
    redraw();
}

static void do_5020(const char *line) {
    if (mode == M_NOTREADY) mode = M_READY;
    sendit("REFRESH\r\n");
    do_logerr(line);
}

static void do_2000(const char *origline) {
    char line[4096];
    strncpy(line, origline, sizeof(line));
    line[sizeof(line)-1] = 0;

    ptrarray<char> names;
    for (char *s=strtok(line, " \t"); s; s=strtok(NULL, " \t")) {
	names.add(s);
    }

    reset_scores();

    // names[0] is the 2000
    for (int i=1; i<names.num(); i++) {
	update_score(names[i], 0);
    }

    dolog("*** Game is starting.");
    if (mode == M_READY) mode = M_BOARD;

    sendit("REFRESH\r\n");
    redraw();
}

static const char *getuser(const char *line) {
    const char *user = strchr(line, ' ');
    if (!user) {
	dolog("*** Server gave us bad 2000-series line: %s", line);
	return NULL;
    }
    while (*user==' ') user++;
    static char usercopy[16];
    memset(usercopy, 0, sizeof(usercopy));
    const char *x = strchr(user, ' ');
    if (!x) {
	strncpy(usercopy, user, sizeof(usercopy)-1);
    }
    else {
	size_t len = x-user;
	if (len>sizeof(usercopy)) len = sizeof(usercopy);
	memcpy(usercopy, user, len);
    }
    usercopy[sizeof(usercopy)-1] = 0;
    return usercopy;
}

static void do_2002(const char *line) {
    const char *user = getuser(line);
    strcpy(whoseturn, user);
    ourturn = (!strcasecmp(user, ourname));
    if (ourturn) {
	beep();
	needapprove=0;
    }
    clearbold();
    redraw();
}

static void do_2003(const char *origline) {
    char line[4096];
    strncpy(line, origline, sizeof(line));
    line[sizeof(line)-1] = 0;
    
    char *v[10];
    int k=0;
    for (char *s=strtok(line, " \t"); s; s=strtok(NULL, " \t")) {
	if (k<10) v[k++] = s;
    }
    /* v[0] is the 2003; v[1] is the user; v[2] is the score */
    for (int i=3; i<k; i++) {
	int x=0, y=0, t, m=0;
	for (char *s=strtok(v[i], ","); s; s=strtok(NULL, ",")) {
	    if (m==0) x=atoi(s);
	    else if (m==1) y=atoi(s);
	    else if (m==2) {
		t = s[0];
		board[x][y].bold = 1;
		board[x][y].tile = t;
		board[x][y].phonytile = '.';
	    }
	    m++;
	}
    }
    dolog("*** %s played (%s points)", v[1], v[2]);
    if (!strcasecmp(v[1], ourname)) {
	assert(ourturn);
    }
    else if (!observer) {
	needapprove = 1;
	beep();
    }
    redraw();
}

static void do_2004(const char *line) {
    const char *user = getuser(line);
    dolog("*** %s accepts the play", user);
    redraw();
}

static void do_2005(const char *line) {
    const char *user = getuser(line);
    dolog("*** %s forfeits the turn", user);
    redraw();
}

static void do_2006(const char *line) {
    const char *user = getuser(line);
    dolog("*** %s withdraws the last play!", user);
    needapprove = 0;
    redraw();
}

static void do_2007_2023(const char *line) {
    static int print_2007s=1;

    const char *user = getuser(line);
    int num = 7;

    if (atoi(line)==2023) {
	print_2007s = 0;
	const char *x = strchr(line, ' ');
	if (x) {
	    while (*x==' ') x++;
	    x = strchr(x, ' ');
	    if (x) num = atoi(x);
	}
    }
    else if (!print_2007s) return;

    dolog("*** %s gets %d new tiles", user, num);
    redraw();
}

static void do_2008(const char *line) {
    const char *user = strchr(line, ' ');
    if (!user) return;
    while (*user==' ') user++;
    const char *deltascore = strchr(user, ' ');
    if (!deltascore) return;
    while (*deltascore==' ') deltascore++;
    const char *score = strchr(deltascore, ' ');
    if (!score) return;
    while (*score==' ') score++;

    const char *realuser = getuser(line);

    int iscore = atoi(score);
    update_score(realuser, iscore);
    redraw();
}

static void do_2009(const char *line) {
    const char *user = strchr(line, ' ');
    if (!user) return;
    while (*user==' ') user++;
    const char *text = strchr(user, ' ');
    if (!text) return;
    text++;  /* only once, allow leading spaces */

    const char *realuser = getuser(line);
    dolog("<%s> %s", realuser, text);
    redraw();
}

static void do_2010(const char *line) {
    const char *user = strchr(line, ' ');
    if (!user) return;
    while (*user==' ') user++;
    const char *text = strchr(user, ' ');
    if (!text) return;
    text++;  /* only once, allow leading spaces */

    const char *realuser = getuser(line);
    dolog("*%s* %s", realuser, text);
    redraw();
}

static void do_2025(const char *line) {
    const char *user = strchr(line, ' ');
    if (!user) return;
    while (*user==' ') user++;
    const char *text = strchr(user, ' ');
    if (!text) return;
    text++;  /* only once, allow leading spaces */

    const char *realuser = getuser(line);
    dolog("* %s %s", realuser, text);
    redraw();
}

static void do_2011(const char *line) {
    const char *user = getuser(line);
    dolog("*** %s has quit (someone take over if they don't come back)", user);
    redraw();
}

static void do_2012(const char *line) {
    const char *ypos = strchr(line, ' ');
    if (!ypos) return;
    while (*ypos==' ') ypos++;
    int y = atoi(ypos);
    const char *tiles = strchr(ypos, ' ');
    if (!tiles) return;
    while (*tiles==' ') tiles++;
    
    update_board_line(y, tiles);
    redraw();
}

static void do_2013(const char *line) {
    const char *rack = strchr(line, ' ');
    if (!rack) rack = "";
    while (*rack==' ') rack++;

    set_rack(rack);
    redraw();
}

static void do_2014(const char *line) {
    const char *text = strchr(line, ' ');
    text++;  // only once, allow leading spaces
    dolog("*** %s", text);
    redraw();
}

static void do_2015(const char *) {
    dolog("*** The server is disconnecting");
    redraw();
}

static void do_2016(const char *line) {
    const char *tile = strchr(line, ' ');
    while (*tile==' ') tile++;
    add_rack(tile[0]);
    redraw();
}

static void do_2017(const char *line) {
    const char *user = getuser(line);
    dolog("*** %s has won!", user);
    redraw();
}

static void do_2018(const char *origline) {
    char line[4096];
    strncpy(line, origline, sizeof(line));
    line[sizeof(line)-1] = 0;
    
    ptrarray<char> v;
    for (char *s=strtok(line, " \t"); s; s=strtok(NULL, " \t")) {
	v.add(s);
    }
    dolog("*** Tie game! Winners are:");
    for (int i=1; i<v.num(); i++) dolog("***    %s", v[i]);
    redraw();
}

static void do_2019(const char *origline) {
    char line[4096];
    strncpy(line, origline, sizeof(line));
    line[sizeof(line)-1] = 0;
    
    char *v[3];
    int k=0;
    for (char *s=strtok(line, " \t"); s; s=strtok(NULL, " \t")) {
	if (k<3) v[k++] = s;
    }
    if (k<3) return;
    // k[0] is the 2019
    char safename[16];
    strncpy(safename, v[2], sizeof(safename));
    safename[sizeof(safename)-1] = 0;

    handle_rename(v[1], safename);
    if (!strcasecmp(ourname, v[1])) {
	strcpy(ourname, safename);
    }
    if (!strcasecmp(whoseturn, v[1])) {
	strcpy(whoseturn, safename);
    }
    // don't post a message, the server does (for historical reasons)
    redraw();
}

static void do_2021(const char *) {
    if (in_game && !observer) return;
    if (observer) {
	observer = 0;
	dolog("*** Exiting observer mode.");
    }
    if (!in_game) {
	in_game = 1;
	dolog("*** Joining game.");
    }
    if (mode == M_READY || mode == M_NOTREADY) mode = M_BOARD;
    sendit("REFRESH\r\n");
    redraw();
}

static void do_2022(const char *) {
    needapprove = 1;
    redraw();
}

static void do_2024(const char *) {
    dolog("*** Server is ready to start another game (or use LEAVE).");
    strcpy(whoseturn, "");
    ourturn = 0;
    observer = 0;
    mode = M_NOTREADY;
    needapprove = 0;
    clearbold();
    redraw();
}

static void do_2026(const char *) {
    dolog("*** BEEP ***");
    beep();
    redraw();
}

static void do_1160(const char *line) {
    const char *stuff = strchr(line, ' ');
    if (!stuff) return;
    while (*stuff==' ') stuff++;

    for (int i=0; i<gamelist.num(); i++) gamelist[i].exists=0;

    char realstuff[4096];
    strncpy(realstuff, stuff, sizeof(realstuff));
    realstuff[sizeof(realstuff)-1] = 0;
    for (char *s = strtok(realstuff, " "); s; s = strtok(NULL, " ")) {
	int n = atoi(s);
	game_exists(n);
	sendit("GINFO %d\r\n", n);
    }
    update_live_games();
    redraw();
}

static void do_1170(const char *line) {
    const char *s = strchr(line, ' ');
    if (!s) return;
    while (*s==' ') s++;
    int gamenum = atoi(s);

    s = strchr(s, ' ');
    if (!s) return;
    while (*s==' ') s++;
    int playing = (*s=='P' || *s=='p');

    s = strchr(s, ' ');
    if (!s) return;
    while (*s==' ') s++;
    int npl = atoi(s);

    s = strchr(s, ' ');
    if (!s) return;
    while (*s==' ') s++;
    int nobs = atoi(s);

    if (gamenum<0 || gamenum>=gamelist.num()) return;
    game_exists(gamenum);
    update_live_games();

    gamelist[gamenum].started = playing;
    gamelist[gamenum].players = npl;
    gamelist[gamenum].observers = nobs;
    redraw();
}

static void do_1180_1190(const char *) {
    dolog("*** Joined game");
    in_game = 1;
    redraw();
}

static void do_1200(const char *) {
    if (old_protocol) return;
    dolog("*** Left game");
    in_game = 0;
    ourturn = 0;
    observer = 0;
    reset_scores();
    mode = M_NOTREADY;
    needapprove = 0;
    redraw();
}

static void do_2027(const char *line) {
    const char *s = strchr(line, ' ');
    if (!s) return;
    int n = atoi(s);
    game_exists(n);
    update_live_games();
    gamelist[n].started = 0;
    gamelist[n].players = 0;
    gamelist[n].observers = 0;
    sendit("GINFO %d\r\n", n);
    redraw();
}

static void do_2028(const char *line) {
    const char *s = strchr(line, ' ');
    if (!s) return;
    int n = atoi(s);
    sendit("GINFO %d\r\n", n);
    redraw();
}

static struct {
    int num;
    void (*func)(const char *line);
} table[] = {
    { 1000, do_1000 },
    { 5000, do_500x },
    { 5001, do_500x },

    { 1010, do_1010 },
    { 1020, do_1020_2001 },
    { 1120, do_1120_2020 },
    { 5020, do_5020 },
    // ignore 1030, 1040, 1050, 1060
    // { 5061, do_5061 },
    // ignore 1070, 1080, 1090, 1100, 1110

    { 1160, do_1160 },
    { 1170, do_1170 },
    { 1180, do_1180_1190 },
    { 1190, do_1180_1190 },
    { 1200, do_1200 },

    { 2000, do_2000 },
    { 2001, do_1020_2001 },
    { 2002, do_2002 },
    { 2003, do_2003 },
    { 2004, do_2004 },
    { 2005, do_2005 },
    { 2006, do_2006 },
    { 2007, do_2007_2023 },
    { 2008, do_2008 },
    { 2009, do_2009 },
    { 2010, do_2010 },
    { 2011, do_2011 },
    { 2012, do_2012 },
    { 2013, do_2013 },
    { 2014, do_2014 },
    { 2015, do_2015 },
    { 2016, do_2016 },
    { 2017, do_2017 },
    { 2018, do_2018 },
    { 2019, do_2019 },
    { 2020, do_1120_2020 },
    { 2021, do_2021 },
    { 2022, do_2022 },
    { 2023, do_2007_2023 },
    { 2024, do_2024 },
    { 2025, do_2025 },
    { 2026, do_2026 },
    { 2027, do_2027 },
    { 2028, do_2028 },
    { 0, NULL }
};

static void netline(const char *line) {
    int n = atoi(line);
    bool found = false;
    for (int i=0; table[i].func && !found; i++) {
	if (table[i].num==n) {
	    found = true;
	    table[i].func(line);
	}
    }
    if (!found && n>=5000 && n<6000) do_logerr(line);
}

static void netdata(const char *inbuf, size_t len) {
    static char buf[8192];
    static size_t bufpos=0;

    memcpy(buf+bufpos, inbuf, len);
    bufpos += len;

    char *s;
    while ((s=(char *)memchr(buf, '\n', bufpos))!=NULL) {
	*s=0;
	if (s!=buf && s[-1]=='\r') s[-1]=0;
	s++;
	netline(buf);
	bufpos = bufpos - (s-buf);
	memmove(buf, s, bufpos);
    }
}

////////////////////////////////////////////////////////////

static void mainloop() {
    bool done=false;
    while (!done) {
	fd_set xset;
	FD_ZERO(&xset);
	FD_SET(0, &xset);
	FD_SET(sock, &xset);
	if (select(sock+1, &xset, NULL, NULL, NULL)<0) {
	    if (errno!=EINTR) sleep(1);
	    redraw();
	    continue;
	}
	if (FD_ISSET(0, &xset)) {
	    char ch;
	    int l = read(0, &ch, 1);
	    if (l<=0) done=1;
	    else raw_keystroke((unsigned char)ch);
	}
	if (FD_ISSET(sock, &xset)) {
	    char buf[4096];
	    int l = read(sock, buf, sizeof(buf));
	    if (l<=0) {
		close(sock);
		sock = -1;
		done = 1;
	    }
	    else {
		netdata(buf, l);
	    }
	}
    }
}

static void setup(const char *server) {
    struct hostent *h = gethostbyname(server);
    if (!h) {
	printf("Server %s not found\n", server);
	exit(1);
    }
    strncpy(ourserver, server, sizeof(ourserver));
    ourserver[sizeof(ourserver)-1] = 0;
    
    struct sockaddr_in sn;
    memset(&sn, 0, sizeof(sn));
    sn.sin_family = AF_INET;
    sn.sin_port = htons(SCRAB_PORT);
    memcpy(&sn.sin_addr, h->h_addr, sizeof(sn.sin_addr));

    sock = socket(AF_INET, SOCK_STREAM, 0);
    if (sock<0) {
	printf("socket: %s\n", strerror(errno));
	exit(1);
    }

    if (connect(sock, (struct sockaddr *)&sn, sizeof(sn))) {
	printf("connect: %s\n", strerror(errno));
	exit(1);
    }
}

int main(int argc, char *argv[]) {
    const char *name = getenv("IRCNICK");
    if (!name) name = getlogin();
    const char *server = "localhost";

    struct timeval tv;
    gettimeofday(&tv, NULL);
    srandom(tv.tv_sec ^ tv.tv_usec);

    int ch;
    while ((ch = getopt(argc, argv, "8bc"))!=-1) {
        switch (ch) {
	  case '8': input8bit = 1; break;
	  case 'b': colorbold = 1; break;
	  case 'c': usecolor = 1; break;
	  default: 
	    printf("Illegal option -%c\n", ch);
	    printf("Usage: scrabtty [-8] [-b] [-c] [username@server-host]\n");
	    printf("    -8: 8-bit character input\n");
	    printf("    -b: Implement bold using colors\n");
	    printf("    -c: Enable color\n");
	    return 1;
	}
    }

    for (int i=optind; i<argc; i++) {
	char *at = strchr(argv[i], '@');
	if (at) {
	    *at++ = 0;
	    /* check lengths so you can use user@ and @host and have it work */
	    if (strlen(at)>0) server = at;
	    if (strlen(argv[i])>0) name = argv[i];
	}
	else if (strchr(argv[i], '.') || i>1) server = argv[i];
	else name = argv[i];
    }

    if (!name) {
	printf("No name specified and the system didn't seem to have a "
	       "username for you.\n");
	printf("Please specify a name on the command line.\n");
	return 0;
    }

    strcpy(ourname, name);
    initboard();
    setup(server);
    mainloop();
    endwin();
    printf("\n");
    return 0;
}
