/*
 * Half-Baked Curses
 *
 * This implements a subset of the curses API in terms of the
 * semi-standard vt100/ANSI.SYS/xterm escape sequences. If you
 * have trouble with your normal curses, you might try this. 
 * Note that it does not optimize output very well; if you don't
 * have a fast link it'll probably be quite maddening.
 */

#include <sys/types.h>
#include <sys/ioctl.h>
#include <termios.h>
#include <unistd.h>
#include <stdlib.h>
#include <signal.h>
#include <stdio.h>

#include "curses.h"

#ifndef __HBC_CURSES
#error "messed up includes"
#endif

/* number of possible color combos */
#define MAX_COMBOS   8

#define CLEAR_CHAR   0x0020      /* no attribute/space */

/* terminal codes */
#define CODE_CLEAR    "\x1b[2J"
#define CODE_CLREOL   "\x1b[K"
#define CODE_ATTROFF  "\x1b[0m"
#define CODE_ATTRBOLD "\x1b[1m"
#define CODE_ATTRDIM  "\x1b[2m"
#define CODE_ATTRREV  "\x1b[7m"


int COLS=0, LINES=0;

struct screen {
    short **data;
    int attr;
    int x,y;
};

/* the user updates virt_screen, we update phys_screen at refresh() time */
static screen phys_screen, virt_screen;

/* temporary line buffer used by refresh */
static short *linebuf;

/* if true, clear physical screen at next refresh */
static int clearphys=1;

/* if true, need to reset termios at next refresh */
static int endwinned=0;

/* saved termios for restore at endwin time */
static struct termios orig_termios;

/* current working termios */
static struct termios our_termios;

/* old signal handlers */
static void (*oldstophandler)(int) = SIG_DFL;
static void (*oldconthandler)(int) = SIG_DFL;
static void (*oldwinchandler)(int) = SIG_DFL;

/* color table */
static int bgs[MAX_COMBOS];
static int fgs[MAX_COMBOS];

////////////////////////////////////////////////////////////

static void getsize() {
    COLS=LINES=0;
#ifdef TIOCGWINSZ
    struct winsize ws;
    if (ioctl(0, TIOCGWINSZ, &ws) >= 0) { 
	// Ultrix (at least) likes to put 0 in here if nobody set it.
	if (ws.ws_col>0) COLS=ws.ws_col; 
	if (ws.ws_row>0) LINES=ws.ws_row; 
    }
#else
#ifdef TIOCGSIZE
    struct ttysize ts;
    if (ioctl(0, TIOCGSIZE, &ts) >= 0) { 
	COLS=ts.ts_cols;
	LINES=ts.ts_lines;
    }
#endif
#endif
    if (LINES==0) {
	const char *p = getenv("LINES");
	if (p) LINES = atoi(p);
    }
    if (COLS==0) {
	const char *q = getenv("COLUMNS");
	if (q) COLS = atoi(q);
    }

    /* last-ditch, assume 80x24 */
    if (LINES==0) LINES=24;
    if (COLS==0) COLS=80;
}

static void clearit(screen *s) {
    for (int y=0; y<LINES; y++) {
	for (int x=0; x<COLS; x++) {
	    s->data[y][x] = CLEAR_CHAR;
	}
    }
    s->x = s->y = 0;
    s->attr = 0;
}

static void alloc_screen(screen *s) {
    s->data = new short* [LINES];
    for (int y=0; y<LINES; y++) {
	s->data[y] = new short[COLS+1];
	s->data[y][COLS]=0;
    }
    clearit(s);
}

static void realloc_screen(screen *s, int oldcols, int oldlines) {
    if (LINES > oldlines) {
       short **newdata = new short* [LINES];
       for (int y=0; y<oldlines; y++) {
	  newdata[y] = s->data[y];
       }
       for (int y=oldlines; y<LINES; y++) {
	  newdata[y] = new short[oldcols+1];
	  for (int x=0; x<oldcols; x++) {
	     newdata[y][x] = CLEAR_CHAR;
	  }
	  newdata[y][oldcols] = 0;
       }
       delete []s->data;
       s->data = newdata;
    }
    else {
       for (int y=LINES; y<oldlines; y++) {
	  delete [] s->data[y];
	  s->data[y] = NULL;
       }
    }

    if (COLS > oldcols) {
       for (int y=0; y<LINES; y++) {
	  short *line = new short[COLS+1];
	  memcpy(line, s->data[y], oldcols*sizeof(short));
	  for (int x=oldcols; x<COLS; x++) {
	     line[x] = CLEAR_CHAR;
	  }
	  line[COLS] = 0;
	  delete [] s->data[y];
	  s->data[y] = line;
       }
    }
    else {
       for (int y=0; y<LINES; y++) {
	  s->data[y][COLS] = 0;
       }
    }
}

static void changesize(void) {
    int oldcols = COLS;
    int oldlines = LINES;
    getsize();
    if (oldcols!=COLS || oldlines!=LINES) {
       realloc_screen(&virt_screen, oldcols, oldlines);
       realloc_screen(&phys_screen, oldcols, oldlines);
    }
}

////////////////////////////////////////////////////////////

static void limit_pos() {
    if (virt_screen.y >= LINES) virt_screen.y = LINES-1;
    if (virt_screen.x >= COLS) virt_screen.x = COLS-1;
    if (virt_screen.y < 0) virt_screen.y = 0;
    if (virt_screen.x < 0) virt_screen.x = 0;
}

void erase() {
    clearit(&virt_screen);
}

void clear() {
    clearphys = 1;
}
 
void clrtoeol() {
    for (int x=virt_screen.x; x<COLS; x++) {
	virt_screen.data[virt_screen.y][x] = CLEAR_CHAR;
    }
}

void move(int y, int x) {
    virt_screen.y = y;
    virt_screen.x = x;
    limit_pos();
}

void addch(int ch) {
    virt_screen.data[virt_screen.y][virt_screen.x++] = virt_screen.attr|ch;
    limit_pos();
}

void addstr(const char *s) {
    while (*s) addch(*(s++));
}

/* should this wait until refresh time? */
void beep() {
    static const char c='\7';
    write(STDOUT_FILENO, &c, 1);
}

void attron(int attr) {
    virt_screen.attr |= attr;
}

void attroff(int attr) {
    virt_screen.attr &= ~attr;
}

void mvaddch(int y, int x, int ch) {
    move(y, x);
    addch(ch);
}

void mvaddstr(int y, int x, const char *s) {
    move(y, x);
    addstr(s);
}

////////////////////////////////////////////////////////////

static void emit(const char *str) {
    fputs(str, stdout);
}

static void emit_move(int y, int x) {
    printf("\x1b[%d;%dH", y+1, x+1);
}

static void emit_color(int combo) {
    int fg = fgs[combo];
    int bg = bgs[combo];
    printf("\x1b[%d;%dm", 30+fg, 40+bg);
}

static void adjust_attr(int attr) {
    if (attr==0) {
	if (phys_screen.attr!=0) emit(CODE_ATTROFF);
    }
    else {
	int curcolor = (phys_screen.attr & 0xF000) >> 12;
	int newcolor = (attr & 0xF000) >> 12;
	int on = (attr & 0x0F00) & ~phys_screen.attr;
	int off = phys_screen.attr & ~(0x0F00 & attr);
	if (off) {
	    // there are actually codes to turn attributes off one at a time,
	    // but stupider terminals don't know them.
	    emit(CODE_ATTROFF);
	    on = attr;
	    curcolor = 0;
	}
	if (curcolor!=newcolor) emit_color(newcolor);
	if (on & A_BOLD) emit(CODE_ATTRBOLD);
	if (on & A_DIM) emit(CODE_ATTRDIM);
	if (on & A_REVERSE) emit(CODE_ATTRREV);
    }
    phys_screen.attr = attr;
}

static void emit_text(const short *txt) {
    for (int i=0; txt[i]; i++) {
	short attr = txt[i] & 0xff00;
	if (phys_screen.attr != attr) {
	    adjust_attr(attr);
	}
	putchar(txt[i]&0xff);
    }
}

////////////////////////////////////////////////////////////

static void dostop(int sig) {
    emit_move(LINES-1, 0);
    if (oldstophandler) oldstophandler(sig);
    endwin();
    kill(getpid(), SIGSTOP);
}

static void docont(int sig) {
    clear();
    refresh();
    if (oldconthandler) oldconthandler(sig);
}

static void dowinch(int sig) {
    clear();
    changesize();
    if (oldwinchandler) oldwinchandler(sig);
    refresh();
}

static void establish_signals() {
    struct sigaction sa, old;
    sa.sa_flags = 0;
    sigemptyset(&sa.sa_mask);

    sa.sa_handler = docont;
    sigaction(SIGCONT, &sa, &old);
    oldconthandler = old.sa_handler;

    sa.sa_handler = dostop;
    sigaction(SIGTSTP, &sa, &old);
    oldstophandler = old.sa_handler;

    sa.sa_handler = dowinch;
    sigaction(SIGWINCH, &sa, &old);
    oldwinchandler = old.sa_handler;
}

////////////////////////////////////////////////////////////

void initscr() {
    tcgetattr(STDIN_FILENO, &orig_termios);
    our_termios = orig_termios;
    clearphys = 1;
    endwinned = 0;
    getsize();

    alloc_screen(&virt_screen);
    alloc_screen(&phys_screen);

    linebuf = new short[COLS+1];

    establish_signals();
}

void cbreak() {
#ifdef XCASE
    our_termios.c_lflag &= ~(XCASE); 
#endif
    our_termios.c_lflag &= ~(ECHONL|NOFLSH); 
//  our_termios.c_lflag &= ~(ISIG); 
    our_termios.c_lflag |= ISIG;
    our_termios.c_lflag &= ~(ICANON); 
    our_termios.c_iflag &= ~(ICRNL | INLCR);
    our_termios.c_cflag |= CREAD;
    our_termios.c_cc[VTIME] = 0;
    our_termios.c_cc[VMIN] = 1;
    tcsetattr(STDIN_FILENO, TCSADRAIN, &our_termios);
}

void noecho() {
    /* grr, glibc is broken */
#ifdef IECHO
    our_termios.c_lflag &= ~(IECHO); 
#endif
#ifdef ECHO
    our_termios.c_lflag &= ~(ECHO); 
#endif
    tcsetattr(STDIN_FILENO, TCSADRAIN, &our_termios);
}

void endwin() {
    tcsetattr(STDIN_FILENO, TCSADRAIN, &orig_termios);
    endwinned=1;
}

void refresh(void) {
    if (endwinned) {
	tcsetattr(STDIN_FILENO, TCSADRAIN, &our_termios);
	changesize();
	endwinned=0;
    }

    if (clearphys) {
	clearit(&phys_screen);
	emit(CODE_CLEAR);
	clearphys = 0;
    }

    for (int y=0; y<LINES; y++) {
	// skip over common prefix, if any
	int k=0;
	while (k<COLS && virt_screen.data[y][k]==phys_screen.data[y][k]) k++;

	// skip identical lines completely
	if (k==COLS) continue;

	emit_move(y, k);

	// copy rest of line to linebuf
	int linebuflen=0;
	for (int x=k; x<=COLS; x++) {
	    linebuf[linebuflen++] = virt_screen.data[y][x];
	}
	linebuf[linebuflen] = 0;

	// wipe out trailing spaces
	for (int i=linebuflen-1; i>=0 && linebuf[i]==CLEAR_CHAR; i--) {
	    linebuf[i] = 0;
	}

	emit_text(linebuf);
	emit(CODE_CLREOL);

	for (int x=0; x<COLS; x++) {
	    phys_screen.data[y][x] = virt_screen.data[y][x];
	}
    }

    phys_screen.x = virt_screen.x;
    phys_screen.y = virt_screen.y;
    emit_move(phys_screen.y, phys_screen.x);

    /* leave screen attributes off in case someone else writes to tty */
    adjust_attr(0);

    fflush(stdout);
}

////////////////////////////////////////////////////////////

bool has_colors() {
    /* Assume user will not request colors if they don't work */
    return 1;
}

void start_color() {
    for (int i=0; i<MAX_COMBOS; i++) {
	bgs[i] = COLOR_BLACK;
	fgs[i] = COLOR_WHITE;
    }
}

void init_pair(int n, int fg, int bg) {
    if (n>=1 && n<MAX_COMBOS) {
	bgs[n] = bg;
	fgs[n] = fg;
    }
}
