
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <errno.h>
#include <ctype.h>
#include <time.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <stdarg.h>
#include <assert.h>

#include "primarray.h"
#include "ptrarray.h"
#include "version.h"

#include "server.h"
#include "tile.h"
#include "tileq.h"
#include "rack.h"
#include "board.h"

struct client {
    char hostname[256];     /* connecting host */
    char buffer[8192];      /* network buffer */
    size_t bufpos;          /* length of buffer presently valid */
    int fd;                 /* socket */
    int playernum;          /* index into players[] */
};

struct player {
    int clientnum;          /* index into clients[] */
    char username[16];      /* username in game */
    int hasname;            /* starts 0, becomes 1 after first NAME */
    time_t lastactivity;    /* when last did anything */

    int gamenum;            /* game number playing in */
    int playernum;          /* play order, starts -1, assigned at start time */
    int ready;              /* ready to play? 0/1 Y/N; -1 = observer */
    int starttile;          /* starting tile assigned in READY */

    int approved;           /* approved current play */
    int forfeited;          /* forfeited last turn */

    rack undorack;          /* previous tiles available */
    rack myrack;            /* tiles available */
    int score;              /* current score */
    int tentativescore;     /* additional unapproved score */
};

struct game {
    int num;

    int game_started;
    primarray<int> ordered_players;  /* in play order; indexes players[] */
    int whoseturn;                   /* index to ordered_players */
    tileq tilequeue;
    
    board undoboard;
    board theboard;
    bool undofirstplay;
    bool firstplay;
    int undo_who;            /* who can undo, only if a play was made */


    game() {
	game_started=0;
	whoseturn=-1;
	undofirstplay = false;
	firstplay = false;
	undo_who=-1;
    }
};

static ptrarray<client> clients;
static ptrarray<player> players;
static ptrarray<game> games;

static int use_dns = 1, reject_no_dns = 1;

static void message(player *p, const char *fmt, ...) {
    if (p->clientnum<0) return;
    assert(clients[p->clientnum]);
    char buf[4096];
    va_list ap;
    va_start(ap, fmt);
    vsnprintf(buf, sizeof(buf), fmt, ap);
    va_end(ap);
    write(clients[p->clientnum]->fd, buf, strlen(buf));
}

static void broadcast(const char *fmt, ...) {
    char buf[4096];
    va_list ap;
    va_start(ap, fmt);
    vsnprintf(buf, sizeof(buf), fmt, ap);
    va_end(ap);
    for (int i=0; i<clients.num(); i++) if (clients[i]) {
	write(clients[i]->fd, buf, strlen(buf));
    }
}

static void gbroadcast(game *g, const char *fmt, ...) {
    char buf[4096];
    va_list ap;
    va_start(ap, fmt);
    vsnprintf(buf, sizeof(buf), fmt, ap);
    va_end(ap);
    for (int i=0; i<clients.num(); i++) if (clients[i]) {
	player *p = players[clients[i]->playernum];
	assert(p);
	if (p->gamenum == g->num) {
	    write(clients[i]->fd, buf, strlen(buf));
	}
    }
}

//////////////////////////////////////////////////////////////////////

static game *game_of_player(player *pl) {
    if (pl->gamenum==0) return NULL;
    return games[pl->gamenum-1];
}

static void player_unidle(player *pl) {
   pl->lastactivity = time(NULL);
}

//////////////////////////////////////////////////////////////////////

static void report_rack(player *c) {
    const rack *r = &c->myrack;
    assert(count_rack(r)<=7);
    char buf[8];
    spill_rack(r, buf);
    message(c, "2013 %s\r\n", buf);
}

//////////////////////////////////////////////////////////////////////

static void draw_tiles(game *g, player *c) {
    tileq *q = &g->tilequeue;
    if (q->tilequeueempty) return;
    while (count_rack(&c->myrack)<7) {
	int tile = picktile(q);
	if (tile==0) break;

	message(c, "2016 %c drawn (for rack)\r\n", tile);
	add_rack(&c->myrack, tile);
    }
    int nleft = strlen(q->tilequeue+q->tilequeuepos);
    if (nleft==0) {
	q->tilequeueempty = true;
	gbroadcast(g, "2014 No more tiles left.\r\n");
    }
    else {
	/* Round to nearest 10. */
	int r = 10*((nleft+5)/10);
	if (r>0 && nleft > 8) {
	    message(c, "2014 There are around %d tiles left.\r\n", r);
	}
	else if (nleft==1) {
	    message(c, "2014 There is 1 tile left.\r\n");
	}
	else {
	    message(c, "2014 There are %d tiles left.\r\n", nleft);
	}
    }
}

//////////////////////////////////////////////////////////////////////

static const board letterscores = {{
    { 1, 1, 1, 2, 1, 1, 1, 1, 1, 1, 1, 2, 1, 1, 1 },
    { 1, 1, 1, 1, 1, 3, 1, 1, 1, 3, 1, 1, 1, 1, 1 },
    { 1, 1, 1, 1, 1, 1, 2, 1, 2, 1, 1, 1, 1, 1, 1 },
    { 2, 1, 1, 1, 1, 1, 1, 2, 1, 1, 1, 1, 1, 1, 2 },
    { 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1 },
    { 1, 3, 1, 1, 1, 3, 1, 1, 1, 3, 1, 1, 1, 3, 1 },
    { 1, 1, 2, 1, 1, 1, 2, 1, 2, 1, 1, 1, 2, 1, 1 },
    { 1, 1, 1, 2, 1, 1, 1, 1, 1, 1, 1, 2, 1, 1, 1 },
    { 1, 1, 2, 1, 1, 1, 2, 1, 2, 1, 1, 1, 2, 1, 1 },
    { 1, 3, 1, 1, 1, 3, 1, 1, 1, 3, 1, 1, 1, 3, 1 },
    { 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1 },
    { 2, 1, 1, 1, 1, 1, 1, 2, 1, 1, 1, 1, 1, 1, 2 },
    { 1, 1, 1, 1, 1, 1, 2, 1, 2, 1, 1, 1, 1, 1, 1 },
    { 1, 1, 1, 1, 1, 3, 1, 1, 1, 3, 1, 1, 1, 1, 1 },
    { 1, 1, 1, 2, 1, 1, 1, 1, 1, 1, 1, 2, 1, 1, 1 },
}};
static const board wordscores = {{
    { 3, 1, 1, 1, 1, 1, 1, 3, 1, 1, 1, 1, 1, 1, 3 },
    { 1, 2, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 2, 1 },
    { 1, 1, 2, 1, 1, 1, 1, 1, 1, 1, 1, 1, 2, 1, 1 },
    { 1, 1, 1, 2, 1, 1, 1, 1, 1, 1, 1, 2, 1, 1, 1 },
    { 1, 1, 1, 1, 2, 1, 1, 1, 1, 1, 2, 1, 1, 1, 1 },
    { 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1 },
    { 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1 },
    { 3, 1, 1, 1, 1, 1, 1, 2, 1, 1, 1, 1, 1, 1, 3 },
    { 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1 },
    { 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1 },
    { 1, 1, 1, 1, 2, 1, 1, 1, 1, 1, 2, 1, 1, 1, 1 },
    { 1, 1, 1, 2, 1, 1, 1, 1, 1, 1, 1, 2, 1, 1, 1 },
    { 1, 1, 2, 1, 1, 1, 1, 1, 1, 1, 1, 1, 2, 1, 1 },
    { 1, 2, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 2, 1 },
    { 3, 1, 1, 1, 1, 1, 1, 3, 1, 1, 1, 1, 1, 1, 3 },
}};


static void reset_board(game *g) {
    for (int i=0; i<BOARDSIZE; i++) for (int j=0; j<BOARDSIZE; j++) {
	g->theboard.v[i][j] = '.';
    }
    g->firstplay = true;
    g->undofirstplay = true;
    g->undo_who = -1;
}

static void send_board(game *g, player *c) {
    message(c, "2999    0123456789 1234\r\n");
    for (int y=0; y<BOARDSIZE; y++) {
	char buf[BOARDSIZE+1];
	for (int x=0; x<BOARDSIZE; x++) {
	    buf[x] = g->theboard.v[x][y];
	}
	buf[BOARDSIZE] = 0;
	message(c, "2012 %-2d %s\r\n", y, buf);
    }
}

static void broadcast_board(game *g) {
    for (int i=0; i<players.num(); i++) {
	if (players[i] && players[i]->gamenum==g->num) {
	    send_board(g, players[i]);
	}
    }
}

static void endgame(game *g, int player_who_ran_out) {
    gbroadcast(g, "2014 Game has ended. Adjusting scores...\r\n");
    int tscore = 0;
    for (int i=0; i<g->ordered_players.num(); i++) {
	player *c = players[g->ordered_players[i]];
	int score = score_rack(&c->myrack);
	if (i!=player_who_ran_out) {
	    c->tentativescore = -score;
	    tscore += score;
	    gbroadcast(g, "2008 %s -%d %d (final score)\r\n", c->username, 
		       score, c->score + c->tentativescore);
	    gbroadcast(g, "2014 %s loses %d %s\r\n", c->username, score,
		       score==1 ? "point" : "points");
	}
	else {
	    assert(score==0);
	}
    }
    if (player_who_ran_out>=0) {
	player *c = players[g->ordered_players[player_who_ran_out]];
	assert(c->tentativescore==0);
	c->tentativescore = tscore;
	gbroadcast(g, "2008 %s %d %d (final score)\r\n", c->username, tscore,
		   c->score + c->tentativescore);
	gbroadcast(g, "2014 %s gains %d %s\r\n", c->username, tscore,
		   tscore==1 ? "point" : "points");
    }

    bool tie=false;
    int winscore = 0;
    int winix=-1;
    primarray<int> extrawinners;
    for (int i=0; i<g->ordered_players.num(); i++) {
	player *c = players[g->ordered_players[i]];
	int sc = c->score+c->tentativescore;
	if (sc > winscore) {
	    tie = false;
	    winscore = sc;
	    winix = i;
	    extrawinners.setsize(0);
	}
	else if (sc==winscore) {
	    tie = true;
	    extrawinners.add(i);
	}
    }

    if (!tie) {
	gbroadcast(g, "2017 %s wins\r\n", 
		   players[g->ordered_players[winix]]->username);
	return;
    }

    /* 
     * Otherwise, report a tie. The official rules call for
     * consulting the unadjusted scores to break ties. In fact,
     * they imply, though do not state, that if the scores are
     * A=302, B=303, and C=303, and the unadjusted scores are
     * A=308, B=304, and C=306, then A wins, even though A was
     * not participating in the tie. I think this is stupid.
     * Furthermore, this mechanism does not tell you what to do
     * if you still have a tie after that. Ties don't happen,
     * anyway.
     */

    extrawinners.add(winix);
    gbroadcast(g, "2099 Tie game; the winners are:\r\n");
    gbroadcast(g, "2018");
    for (int i=0; i<extrawinners.num(); i++) {
	gbroadcast(g, " %s", 
		   players[g->ordered_players[extrawinners[i]]]->username);
    }
    gbroadcast(g, "\r\n");
}

static void reset_player(player *p) {
    p->ready = 0;
    p->starttile = 0;

    p->approved = 0;
    p->forfeited = 0;

    init_rack(&p->undorack);
    init_rack(&p->myrack);
}

static void gamereset(game *g) {
    g->game_started = 0;
    
    for (int i=0; i<players.num(); i++) {
	if (players[i] && players[i]->gamenum == g->num) {
	    if (players[i]->clientnum<0) {
		// drop disconnected guys
		delete players[i];
		players[i] = NULL;
	    }
	    else {
		reset_player(players[i]);
	    }
	}
    }

    randomize_tiles(&g->tilequeue);

    gbroadcast(g, "2024 This game reset - signal when ready to start "
	       "new game.\r\n");
}

static void checkreset(game *g) {
    if (!g->game_started) return;

    for (int i=0; i<players.num(); i++) {
	if (players[i] && players[i]->gamenum==g->num) {
	    if (players[i]->clientnum>=0 && players[i]->ready>0) return;
	}
    }

    // No connected players in the game. Punt it.
    printf("All players disconnected -- game %d reset.\n", g->num);
    gamereset(g);
}

static void nextturn(game *g) {
    g->whoseturn = (g->whoseturn+1) % g->ordered_players.num();
    player *c = players[g->ordered_players[g->whoseturn]];
    gbroadcast(g, "2002 %s (player's turn)\r\n", c->username);
    c->forfeited = 0;
}

/* Returns true if each place in p[] is also listed in word[]. */
static bool play_contains(const primarray<pos> &word, const play *p, int np) {
   for (int i=0; i<np; i++) {
      bool found = false;
      for (int j=0; j<word.num() && !found; j++) {
	 if (word[j].x==p[i].x && word[j].y==p[i].y) found = true;
      }
      if (!found) return false;
   }
   return true;
}

static bool find_word(const board &b, primarray<pos> &word, play pp, pos d) {
    bool legal = false;
    pos p(pp.x, pp.y);
    pos p0(p);
    word.setsize(0);

    /* back up while we have room and letters */
    while (p.x>=0 && p.y>=0 && b.v[p.x][p.y]!='.') {
	p0 = p;
	p -= d;
    }

    /* now roll forward adding the positions */
    for (p=p0; p.x<BOARDSIZE && p.y<BOARDSIZE && b.v[p.x][p.y]!='.'; p+=d) {
	if (tile_isname(b.v[p.x][p.y])) {
	    legal=true;
	}
	word.add(p);
    }
    
    /* return if we found a letter that wasn't ours */
    return legal;
}

static int score_word(const board &b, const primarray<pos> &word) {
    int score=0;
    int wmults=1;

    for (int i=0; i<word.num(); i++) {
	int tile = b.v[word[i].x][word[i].y];
	int lmult = 1;

	/* Get multipliers (double/triple word or letter scores) */
	if (tile_isspec(tile)) {
	    /* one we put down, eligible for multipliers */
	    lmult = letterscores.v[word[i].x][word[i].y];
	    wmults *= wordscores.v[word[i].x][word[i].y];
	}

	/* Get score for this letter */
	score += (tilescore(tile)*lmult);
    }

    score *= wmults;

    return score;
}


static void clientplayed(game *g, player *c, const play *places, int nplaces, 
			 bool horiz)
{
    rack r;
    init_rack(&r);
    for (int i=0; i<nplaces; i++) add_rack(&r, places[i].tile);
    if (!subsetof_rack(&c->myrack, &r)) {
	message(c, "5031 You don't have those tiles.\r\n");
	return;
    }

    pos dp(0,0);          /* delta-position */
    pos pdp(0,0);         /* perpendicular-delta-position */
    if (horiz) dp.x=pdp.y=1;
    else dp.y=pdp.x=1;

    /*
     * Put the letters down, in lowercase at first so we know which they are
     * when scoring.
     */
    board tmpb = g->theboard;

    for (int i=0; i<nplaces; i++) {
	if (tmpb.v[places[i].x][places[i].y]!='.') {
	    message(c, "5032 You may not play over existing tiles.\r\n");
	    return;
	}
	int ch = tile_name2spec(places[i].tile);
	tmpb.v[places[i].x][places[i].y] = ch;
    }

    /*
     * First: work out the whole word in the direction we're going.
     * Note that, while "places" contains it, it is not necessarily
     * the whole thing.
     */
    bool legal=false;
    primarray<pos> word;
    int score=0;

    legal = find_word(tmpb, word, places[0], dp) || legal;
    if (word.num()>1 || g->firstplay) score += score_word(tmpb, word);

    /*
     * If places[] is not a subset of word[], there were gaps in the play.
     * This is not permitted.
     */
    if (!play_contains(word, places, nplaces)) {
       message(c, "5036 Your play may not have gaps in it.\r\n");
       return;
    }

    /*
     * Now, score all the crossing words, if any.
     */
    for (int i=0; i<nplaces; i++) {
	legal = find_word(tmpb, word, places[i], pdp) || legal;
	if (word.num()>1) score += score_word(tmpb, word);
    }

    if (nplaces==7) score += 50;

    g->undofirstplay = false;
    if (!legal) {
	if (g->firstplay) {
	    for (int i=0; i<nplaces; i++) {
		if (places[i].x==BOARDSIZE/2 && places[i].y==BOARDSIZE/2) {
		    legal = true;
		}
	    }
	    if (legal) {
		g->undofirstplay = true;
		g->firstplay = false;
	    }
	    else {
		message(c, "5030 You must start in the middle.\r\n");
		return;
	    }
	}
	else {
	    message(c, "5030 You must connect to existing tiles.\r\n");
	    return;
	}
    }

    /*
     * Uppercase the lowercase tiles.
     */
    for (int i=0; i<nplaces; i++) {
	int ch = tmpb.v[places[i].x][places[i].y];
	ch = tile_spec2name(ch);
	tmpb.v[places[i].x][places[i].y] = ch;
    }

    /*
     * Ok, play is legal (at least on technical grounds), commit it.
     */

    g->undoboard = g->theboard;
    g->undo_who = c->playernum;
    g->theboard = tmpb;
    c->undorack = c->myrack;

    message(c, "1030 Done.\r\n");

    gbroadcast(g, "2003 %s %d ", c->username, score);
    for (int i=0; i<nplaces; i++) {
	gbroadcast(g, "%d,%d,%c ", places[i].x, places[i].y, places[i].tile);
    }
    gbroadcast(g, "(tiles played)\r\n");

    c->tentativescore = score;

    gbroadcast(g, "2008 %s %d %d (score updated)\r\n", c->username, 
	       score, c->score+score);

    subtract_rack(&c->myrack, &r);
    report_rack(c);

    for (int i=0; i<g->ordered_players.num(); i++) {
	players[g->ordered_players[i]]->approved = 0;
    }
    c->approved = 1;
}

static void checkapproval(game *g) {
    bool ok = true;
    for (int i=0; i<g->ordered_players.num(); i++) {
	if (!players[g->ordered_players[i]]->approved) ok = false;
    }
    if (!ok) return;

    player *c = players[g->ordered_players[g->undo_who]];
    assert(c);
    draw_tiles(g, c);
    c->score += c->tentativescore;
    c->tentativescore = 0;
    g->undo_who = -1;

    int ntiles = count_rack(&c->myrack);

    if (g->tilequeue.tilequeueempty && ntiles==0) {
	endgame(g, c->playernum);
	gamereset(g);
    }
    else {
	if (g->tilequeue.tilequeueempty) {
	    gbroadcast(g, "2014 %s has %d %s left.\r\n", c->username, ntiles,
		       ntiles==1 ? "tile" : "tiles");
	}
	nextturn(g);
    }
}

static int checkforfeits(game *g) {
    if (!g->tilequeue.tilequeueempty) return 0;

    bool all=true;
    for (int i=0; i<g->ordered_players.num(); i++) {
	if (!players[g->ordered_players[i]]->forfeited) all = false;
    }
    if (!all) return 0;

    endgame(g, -1);
    gamereset(g);
    return 1;
}

//////////////////////////////////////////////////////////////////////

/* this is part of setting up the play order. */
static int choose_player(tileq *q, const primarray<int> &array, bool draw=false) {
    primarray<int> tiles;
    tiles.setsize(array.num());
    for (int i=0; i<array.num(); i++) {
	if (draw) {
	    tiles[i] = picktile(q);
	    if (tiles[i]==0) {
		/* 
		 * Ran out of tiles while choosing start order. 
		 * This should not happen, but I suppose it's _possible_...
		 */
		randomize_tiles(q);
	    }
	    message(players[array[i]], "2001 %c drawn\r\n", tiles[i]);
	}
	else {
	    tiles[i] = players[array[i]]->starttile;
	}
    }

    /* 
     * Now we just need to find the highest tile in the list. 
     * Note that _ compares correctly, that is, blanks come in last.
     * At least, I think they do. If not, well, it doesn't matter much.
     */
    bool tie=false;
    int highix=-1;
    int hightile='~';
    for (int i=0; i<tiles.num(); i++) {
	printf("high %c got %c tie %s\n", hightile, tiles[i], tie ? "t" : "f");
	if (tiles[i]==hightile) {
	    tie=true;
	}
	else if (tiles[i] < hightile) {
	    tie=false;
	    hightile = tiles[i];
	    highix = i;
	}
    }

    // Easy case, no tie.
    if (!tie) return highix;

    // Hard case; find the tying guys and recurse, drawing more tiles.
    primarray<int> tiers;
    for (int i=0; i<tiles.num(); i++) {
	if (tiles[i]==hightile) tiers.add(array[i]);
    }
    assert(tiers.num()>1);
    return choose_player(q, tiers, true);
}

static void checkstart(game *g) {
    assert(!g->game_started);

    int ready=1;
    primarray<int> player_map;

    for (int i=0; i<players.num(); i++) {
	if (players[i] && players[i]->gamenum == g->num) {
	    if (players[i]->ready==0) ready=0;
	    if (players[i]->ready>0) player_map.add(i);
	}
    }

    // Require at least two players.
    if (player_map.num()<2) return;

    if (!ready) return;

    /*
     * The following code sets up the game.
     */

    /*
     * Establish play order.
     * Choose the player with the highest tile until all players are assigned.
     * choose_player() handles ties.
     */
    g->ordered_players.setsize(0);

    while (player_map.num()>0) {
	/* choose_player returns an index into the array */
	int p = choose_player(&g->tilequeue, player_map);
	/* get the index into clients[] */
	int pn = player_map[p];
	/* save it */
	g->ordered_players.add(pn);

	/* remove the player from player_map */
	int k = player_map.num();
	if (p<k-1) player_map[p] = player_map[k-1];
	player_map.setsize(k-1);
    }

    broadcast("2028 %d gamelist info changed\r\n", g->num);
    gbroadcast(g, "2999 Game is starting.\r\n");
    gbroadcast(g, "2000");
    for (int i=0; i<g->ordered_players.num(); i++) {
	gbroadcast(g, " %s", players[g->ordered_players[i]]->username);
	players[g->ordered_players[i]]->playernum = i;
	players[g->ordered_players[i]]->score = 0;
	players[g->ordered_players[i]]->tentativescore = 0;
    }
    gbroadcast(g, "\r\n");
    g->game_started = 1;

    /* set up game stuff */
    reset_board(g);
    randomize_tiles(&g->tilequeue);

    /* dole out tiles */
    for (int i=0; i<g->ordered_players.num(); i++) {
	draw_tiles(g, players[g->ordered_players[i]]);
    }

    /*
     * Kick off the game by declaring it's the first player's turn.
     */
    g->whoseturn = -1;
    nextturn(g);
}

//////////////////////////////////////////////////////////////////////

static void report_scores(game *g, player *pl) {
    for (int i=0; i<g->ordered_players.num(); i++) {
	player *pl2 = players[g->ordered_players[i]];
	message(pl, "2008 %s 0 %d (score)\r\n", pl2->username, 
		pl2->score + pl2->tentativescore);
    }
}

static void broadcast_scores(game *g) {
    for (int i=0; i<players.num(); i++) {
	if (players[i] && players[i]->gamenum==g->num) {
	    report_scores(g, players[i]);
	}
    }
}

static const char *getstupidusername() {
    static int counter=1;
    static char buf[16];
    snprintf(buf, sizeof(buf), "client-%d", counter++);
    return buf;
}

static int newclient(int newsock, const struct sockaddr_in *sn) {
    struct hostent *h = NULL;

    if (use_dns) {
       h = gethostbyaddr((const char *)&sn->sin_addr,
			 sizeof(sn->sin_addr), AF_INET);
       if (!h && reject_no_dns) {
	  char *msg = "5000 You have no hostname. Access denied.\r\n";
	  write(newsock, msg, strlen(msg));
	  printf("New client: gethostbyaddr failed (closing)\n");
	  return -1;
       }
    }

    const char *hostname = h ? h->h_name : inet_ntoa(sn->sin_addr);

    // make a new client
    client *c = new client;
    strncpy(c->hostname, hostname, sizeof(c->hostname));
    c->hostname[sizeof(c->hostname)-1] = 0;
    for (int i=0; c->hostname[i]; i++) {
	// no funnies in hostnames
	if (c->hostname[i]<33 || c->hostname[i]>126) c->hostname[i]='?';
    }
    printf("New client: connection from %s\n", c->hostname);

    c->bufpos = 0;
    c->fd = newsock;

    // make a new player
    player *p = new player;

    strcpy(p->username, getstupidusername());
    p->hasname = 0;

    p->lastactivity = time(NULL);

    p->gamenum = 0;

    p->score = 0;
    p->tentativescore = 0;

    reset_player(p);

    // add to tables
    int cnum=-1, pnum=-1;

    for (int i=0; i<clients.num() && cnum<0; i++) {
	if (clients[i]==NULL) cnum = i;
    }
    for (int i=0; i<players.num() && pnum<0; i++) {
	if (players[i]==NULL) pnum = i;
    }

    if (cnum<0) cnum = clients.add(c);
    else clients[cnum] = c;
    if (pnum<0) pnum = players.add(p);
    else players[pnum] = p;

    p->clientnum = cnum;
    c->playernum = pnum;

    // send banner
    message(p, "1000 %d Game server %s ready\r\n", PROTO_VER, VERSION);

//    if (games[0]->game_started) {
//	message(p, "2020 Game is already in progress.\r\n");
//    }

    return 0;
}

static void deadclient(client *c) {
    player *p = players[c->playernum];
    if (!p) return;  /* shouldn't happen, but... */

    game *g = game_of_player(p);

    if (!g || !g->game_started || p->ready<=0) {
	/*
	 * Either the game hasn't started yet, or this guy isn't playing.
	 * Just forget about him.
	 */
	broadcast("2014 %s has disconnected.\r\n", p->username);
	players[c->playernum] = NULL;
	delete p;

	/* If the game hasn't started, check if everybody else is ready. */
	if (g && !g->game_started) checkstart(g);

	if (g) broadcast("2028 %d gamelist info changed\r\n", g->num);
    }
    else {
	/* 
	 * The game has started, and we were playing. Leave the player
	 * record for another client (someone else, or the original guy
	 * reconnecting) to pick up.
	 */
	p->clientnum = -1;
	broadcast("2011 %s [game %d] quit, someone take over\r\n", 
		  p->username, g->num);
    }
}

static int check_name(const char *name) {
    static const char *const validchars = 
	"abcdefghijklmnopqrstuvwxyz"
	"ABCDEFGHIJKLMNOPQRSTUVWXYZ"
	"0123456789"
	"-_\\/";
    for (int i=0; name[i]; i++) {
	if (!strchr(validchars, name[i])) return -1;
    }
    return 0;
}

static int name_in_use(const char *name) {
    for (int i=0; i<players.num(); i++) if (players[i]) {
	if (!strcasecmp(players[i]->username, name)) return -1;
    }
    return 0;
}

static int get_takeover_name(const char *name) {
    for (int i=0; i<players.num(); i++) if (players[i]) {
	if (!strcasecmp(players[i]->username,name) && players[i]->clientnum<0){
	    return i;
	}
    }
    return -1;
}

/* sort play structures by the coordinate in which they differ */
static int play_sortfunc(const void *a, const void *b) {
    const play *aa = (const play *)a;
    const play *bb = (const play *)b;
    if (aa->y==bb->y) return aa->x - bb->x;
    return aa->y - bb->y;
}

static int clientline(client *cc, player *pl, char *line) {

    ////////////////////
    // NAME

    if (!strncasecmp(line, "NAME ", 5)) {
        player_unidle(pl);
	char *name = line+5;
	if (strlen(name) > sizeof(pl->username)-1 || check_name(name)) {
	    message(pl, "5010 Illegal name\r\n");
	    return 0;
	}
	int newpnum = get_takeover_name(name);
	if (newpnum>=0 && pl->ready<=0) {
	    // it's a takeover target, and we aren't playing.

	    // Take our existing player entry out of the table.
	    players[cc->playernum] = NULL;
	    // Plug the new player entry into our client entry.
	    cc->playernum = newpnum;
	    player *newpl = players[newpnum];
	    assert(newpl);
	    newpl->clientnum = pl->clientnum;

	    game *g = game_of_player(newpl);

	    if (!pl->hasname) {
		broadcast("2014 %s has reconnected from %s\r\n", name, 
			  cc->hostname);
	    }
	    else {
		broadcast("2019 %s %s name change\r\n", pl->username, name);
		if (g) {
		    broadcast("2014 %s has taken over for %s [game %d]\r\n", 
			      pl->username, name, g->num);
		}
		else {
		    broadcast("2014 %s has taken over for %s\r\n", 
			      pl->username, name);
		}
	    }

	    strcpy(newpl->username, name);
	    message(newpl, "1010 Name set.\r\n");

	    message(newpl, "2021 Took over for %s.\r\n", name);

	    if (g && g->undo_who>=0 && 
		g->undo_who!=newpl->playernum && 
		!newpl->approved)
	    {
		message(newpl, "2022 Approval required.\r\n");
	    }

	    // Delete player struct "pl".
	    delete pl;
	}
	else if (name_in_use(name) && strcasecmp(name, pl->username)) {
	    message(pl, "5011 Name is already in use\r\n");
	}
	else {
	    if (!pl->hasname) {
		pl->hasname = 1;
		broadcast("2014 %s (from %s) has joined\r\n", 
			  name, cc->hostname);
	    }
	    else {
		broadcast("2019 %s %s name change\r\n", pl->username, name);
		broadcast("2014 %s is now known as %s\r\n",pl->username, name);
	    }
	    strcpy(pl->username, name);
	    message(pl, "1010 Name set.\r\n");
	}
    }

    ////////////////////
    // READY

    else if (!strcasecmp(line, "READY")) {
        player_unidle(pl);
	if (!pl->hasname) {
	    message(pl, "5023 Use NAME first.\r\n");
	    return 0;
	}
	game *g = game_of_player(pl);
	if (!g) {
	    message(pl, "5022 You're not in a game.\r\n");
	}
	else if (g->game_started) {
	    message(pl, "5020 Game has already started.\r\n");
	}
	else if (pl->ready<=0) {
	    pl->ready = 1;
	    pl->starttile = picktile(&g->tilequeue);
	    message(pl, "1020 %c Your starting tile\r\n", pl->starttile);
	    gbroadcast(g, "2014 %s is ready to start.\r\n", pl->username);
	    checkstart(g);
	}
	else {
	    message(pl, "5021 You already signaled ready.\r\n");
	}
    }

    ////////////////////
    // PLAY

    else if (!strncasecmp(line, "PLAY ", 5)) {
        player_unidle(pl);
	if (!pl->hasname) {
	    message(pl, "5035 Use NAME first.\r\n");
	    return 0;
	}
	game *g = game_of_player(pl);
	if (!g) {
	    message(pl, "5034 Not in a game.\r\n");
	    return 0;
	}
	if (!g->game_started || pl->playernum!=g->whoseturn || g->undo_who>=0){
	    message(pl, "5033 Not your turn.\r\n");
	    return 0;
	}
	char *v[8];
	play ply[7];
	int k=0;
	for (char *s=strtok(line, " \t"); s; s=strtok(NULL, " \t")) {
	    if (k<8) v[k++] = s;
	}
	for (int i=1; i<k; i++) {
	    int m=0;
	    for (char *s=strtok(v[i], ","); s; s=strtok(NULL, ",")) {
		if (m==0) ply[i-1].x = atoi(s);
		else if (m==1) ply[i-1].y = atoi(s);
		else if (m==2) ply[i-1].tile = toupper(s[0]);
		m++;
	    }
	}
	k--;
	/* at this point we have k letters, ply[0..k], k<=7 */
	if (k<0) {
	    message(pl, "5999 Invalid PLAY directive\r\n");
	    return 0;
	}

	bool horiz=true, vert=true;
	for (int i=0; i<k; i++) {
	    if (ply[i].x<0 || ply[i].x>=BOARDSIZE ||
		ply[i].y<0 || ply[i].y>=BOARDSIZE ||
		!tile_isname(ply[i].tile)) 
	    {
		message(pl, "5999 Invalid PLAY directive\r\n");
		return 0;
	    }
	    if (ply[i].x!=ply[0].x) vert=false;
	    if (ply[i].y!=ply[0].y) horiz=false;
	}

	if (!horiz && !vert) {
	    message(pl, "5030 Your play may not turn corners.\r\n");
	    return 0;
	}

	// note: legit for horiz and vert to both be true - that's
	// the case for a 1-letter play. Logic in clientplayed()
	// tries to prevent scoring twice (that is, once each direction)
	// on these.

	qsort(ply, k, sizeof(ply[0]), play_sortfunc);
	
	clientplayed(g, pl, ply, k, horiz);
    }

    ////////////////////
    // APPROVE

    else if (!strcasecmp(line, "APPROVE") || !strcasecmp(line, "ACCEPT")) {
        player_unidle(pl);
	if (!pl->hasname) {
	    message(pl, "5043 Use NAME first.\r\n");
	    return 0;
	}
	game *g = game_of_player(pl);
	if (!g) {
	    message(pl, "5042 Not in a game.\r\n");
	    return 0;
	}
	if (g->undo_who<0 || pl->ready<=0) {
	    message(pl, "5040 No approval required right now.\r\n");
	    return 0;
	}
	if (g->undo_who==pl->playernum) {
	    message(pl, "5041 You can't approve your own word.\r\n");
	    return 0;
	}
	message(pl, "1040 Ok.\r\n");
	gbroadcast(g, "2004 %s accepts that.\r\n", pl->username);
	pl->approved = 1;
	checkapproval(g);
    }

    ////////////////////
    // WITHDRAW

    else if (!strcasecmp(line, "WITHDRAW")) {
        player_unidle(pl);
	if (!pl->hasname) {
	    message(pl, "5052 Use NAME first.\r\n");
	    return 0;
	}
	game *g = game_of_player(pl);
	if (!g) {
	    message(pl, "5051 Not in a game.\r\n");
	    return 0;
	}

	if (!g->game_started || g->undo_who<0 || g->undo_who!=pl->playernum || 
	    pl->playernum!=g->whoseturn) 
	{
	    message(pl, "5050 No play to withdraw\r\n");
	    return 0;
	}
	message(pl, "1050 Move withdrawn, try again\r\n");
	gbroadcast(g, "2006 %s withdraws the last play.\r\n", pl->username);
	g->theboard = g->undoboard;
	g->firstplay = g->undofirstplay;
	g->undo_who = -1;
	pl->tentativescore = 0;
	pl->myrack = pl->undorack;
	broadcast_board(g);
	report_rack(pl);
	broadcast_scores(g);
    }

    ////////////////////
    // EXCHANGE

    else if (!strncasecmp(line, "EXCHANGE ", 9)) {
        player_unidle(pl);
	if (!pl->hasname) {
	    message(pl, "5064 Use NAME first.\r\n");
	    return 0;
	}
	game *g = game_of_player(pl);
	if (!g) {
	    message(pl, "5063 Not in a game.\r\n");
	    return 0;
	}
	if (!g->game_started || pl->playernum!=g->whoseturn) {
	    message(pl, "5060 Not your turn.\r\n");
	    return 0;
	}
	if (count_rack(&pl->myrack)<7 || strlen(g->tilequeue.tilequeue+g->tilequeue.tilequeuepos)<7) {
	    message(pl, "5061 No exchanges this late in the game.\r\n");
	    return 0;
	}

	rack losers;            // letters to discard
	init_rack(&losers);

	char *x = strchr(line, ' ');
	if (x) {
	    while (*x==' ') x++;
	    if (strlen(x)==0) x = NULL;
	}

	if (!x) {
	    message(pl, "5999 Invalid EXCHANGE - must specify tiles.\r\n");
	    return 0;
	}

	while (*x!=0 && *x!=' ' && *x!='\t') {
	    int ch = toupper(*x);
	    if (!tile_isname(ch)) {
		message(pl, "5062 Invalid tile.\r\n");
		return 0;
	    }
	    add_rack(&losers, ch);
	    x++;
	}
	assert(count_rack(&losers)>0);
	if (!subsetof_rack(&pl->myrack, &losers)) {
	    message(pl, "5062 You don't have those letters.\r\n");
	    return 0;
	}
	subtract_rack(&pl->myrack, &losers);

	int num = count_rack(&losers);

	message(pl, "1060 Ok.\r\n");
	draw_tiles(g, pl);
	return_tiles_to_queue(&g->tilequeue, &losers);
	report_rack(pl);

	gbroadcast(g, "2023 %s %d letters exchanged.\r\n", pl->username, num);
	//gbroadcast(g, "2007 %s exchanges %d tiles.\r\n", pl->username, num);

	nextturn(g);
    }

    ////////////////////
    // SAY

    else if (!strncasecmp(line, "SAY ", 4)) {
        player_unidle(pl);
	if (!pl->hasname) {
	    message(pl, "5070 Use NAME first.\r\n");
	    return 0;
	}
	const char *text = line+4;
	broadcast("2009 %s %s\r\n", pl->username, text);
	message(pl, "1070 Said.\r\n");
    }

    ////////////////////
    // TELL

    else if (!strncasecmp(line, "TELL ", 5)) {
        player_unidle(pl);
	if (!pl->hasname) {
	    message(pl, "5081 Use NAME first.\r\n");
	    return 0;
	}
	char *user = line+5;
	char *text = strchr(user, ' ');
	if (!text) {
	    message(pl, "5999 Invalid TELL\r\n");
	    return 0;
	}
	*text++ = 0;
	bool found = false;
	for (int i=0; i<players.num() && !found; i++) if (players[i]) {
	    if (!strcasecmp(players[i]->username, user)) {
		message(players[i], "2010 %s %s\r\n", pl->username, text);
		found = true;
	    }
	}
	if (found) message(pl, "1080 Said.\r\n");
	else message(pl, "5080 No such user.\r\n");
    }

    ////////////////////
    // REFRESH

    else if (!strcasecmp(line, "REFRESH")) {
	if (!pl->hasname) {
	    message(pl, "5092 Use NAME first.\r\n");
	    return 0;
	}
	game *g = game_of_player(pl);
	if (!g) {
	    message(pl, "5091 Not in a game.\r\n");
	    return 0;
	}
	if (!g->game_started) {
	    message(pl, "5090 Game not started\r\n");
	    return 0;
	}
	message(pl, "1090 Ok\r\n");
	send_board(g, pl);
	report_rack(pl);
	report_scores(g, pl);
	message(pl, "2002 %s (whose turn)\r\n", 
		players[g->ordered_players[g->whoseturn]]->username);
	message(pl, "2999 Your name is %s\r\n", pl->username);
    }

    ////////////////////
    // QUIT

    else if (!strcasecmp(line, "QUIT")) {
	message(pl, "1100 Ok, disconnecting.\r\n");
	return 1;
    }

    ////////////////////
    // FORFEIT

    else if (!strcasecmp(line, "FORFEIT")) {
        player_unidle(pl);
	if (!pl->hasname) {
	    message(pl, "5112 Use NAME first.\r\n");
	    return 0;
	}
	game *g = game_of_player(pl);
	if (!g) {
	    message(pl, "5111 Not in a game.\r\n");
	    return 0;
	}
	if (!g->game_started || pl->playernum!=g->whoseturn) {
	    message(pl, "5110 Not your turn.\r\n");
	    return 0;
	}
	message(pl, "1110 Ok\r\n");
	gbroadcast(g, "2005 %s forfeits turn.\r\n", pl->username);
	pl->forfeited = 1;

	if (!checkforfeits(g)) nextturn(g);
    }

    ////////////////////
    // OBSERVER

    else if (!strcasecmp(line, "OBSERVER")) {
        player_unidle(pl);
	if (!pl->hasname) {
	    message(pl, "5122 Use NAME first.\r\n");
	    return 0;
	}
	game *g = game_of_player(pl);
	if (!g) {
	    message(pl, "5121 Not in a game.\r\n");
	    return 0;
	}
	if (g->game_started) {
	    if (pl->ready>0) {
		message(pl,"5120 You are presently playing.\r\n");
		return 0;
	    }
	    message(pl, "1120 Ok\r\n");
	}
	else {
	    message(pl, "1120 Ok\r\n");
	    gbroadcast(g, "2014 %s is observing the game.\r\n", pl->username);
	    pl->ready = -1;
	    checkstart(g);
	}
	broadcast("2028 %d gamelist info changed\r\n", g->num);
    }

    ////////////////////
    // WHO

    else if (!strcasecmp(line, "WHO")) {
	if (!pl->hasname) {
	    message(pl, "5130 Use NAME first.\r\n");
	    return 0;
	}
	message(pl, "1130 Ok\r\n");

	int ct = 0;
	for (int i=0; i<players.num(); i++) if (players[i]) ct++;
	assert(ct>0);  // if it were 0, nobody could have sent the WHO.
	message(pl, "2014 %d client%s connected:\r\n", ct, ct==1 ? "" : "s");

	for (int i=0; i<players.num(); i++) if (players[i]) {
	    const char *username = "<anonymous>";
	    if (players[i]->hasname) username = players[i]->username;
	    const char *hostname = "<disconnected>";
	    int cn = players[i]->clientnum;
	    if (cn>=0) hostname = clients[cn]->hostname;
	    char gamedesc[32];
	    if (players[i]->gamenum==0) {
		strcpy(gamedesc, "no game");
	    }
	    else {
		game *g = game_of_player(players[i]);
		assert(g);
		const char *how = NULL;
		switch (players[i]->ready) {
		    case -1: how = "observing"; break;
		    case 0: how = "not ready in"; break;
		    case 1: how = g->game_started ? "playing" : "ready in";
			break;
		}
		snprintf(gamedesc, sizeof(gamedesc), "%s game %d", how,g->num);
	    }
	    char idlebuf[32];
	    time_t now = time(NULL);
	    long idlesecs = (long)(now - players[i]->lastactivity);
	    if (idlesecs > 7200) {
	       snprintf(idlebuf, sizeof(idlebuf), 
			" idle %ld hours", idlesecs/3600);
	    }
	    else if (idlesecs >= 3600) {
	       snprintf(idlebuf, sizeof(idlebuf), 
			" idle %ld:%02ld", idlesecs/3600, (idlesecs/60)%60);
	    }
	    else if (idlesecs >= 60) {
	       snprintf(idlebuf, sizeof(idlebuf),
			" idle %ld minutes", idlesecs/60);
	    }
	    else {
	       *idlebuf = 0;
	    }
	    message(pl, "2014   %s (%s) [%s]%s\r\n", username, hostname, 
		    gamedesc, idlebuf);
	}
    }

    ////////////////////
    // EMOTE

    else if (!strncasecmp(line, "EMOTE ", 6)) {
        player_unidle(pl);
	if (!pl->hasname) {
	    message(pl, "5140 Use NAME first.\r\n");
	    return 0;
	}
	const char *text = line+6;
	broadcast("2025 %s %s\r\n", pl->username, text);
	message(pl, "1140 Emoted.\r\n");
    }

    ////////////////////
    // BEEP

    else if (!strncasecmp(line, "BEEP ", 5)) {
        player_unidle(pl);
	if (!pl->hasname) {
	    message(pl, "5151 Use NAME first.\r\n");
	    return 0;
	}
	char *user = line+5;
	bool found = false;
	for (int i=0; i<players.num() && !found; i++) if (players[i]) {
	    if (!strcasecmp(players[i]->username, user)) {
		message(players[i], "2026 \ahello? (from %s)\r\n", 
			pl->username);
		found = true;
	    }
	}
	if (found) message(pl, "1150 Beeped.\r\n");
	else message(pl, "5150 No such user.\r\n");
    }

    ////////////////////
    // GLIST

    else if (!strcasecmp(line, "GLIST")) {
	message(pl, "2999 Games:\r\n");
	message(pl, "1160 ");
	for (int i=0; i<games.num(); i++) if (games[i]) {
	    message(pl, "%d ", games[i]->num);
	}
	message(pl, "\r\n");
    }

    ////////////////////
    // GINFO

    else if (!strncasecmp(line, "GINFO ", 6)) {
	char *tmp = line+6;
	int gamenum = atoi(tmp);
	int gameix = gamenum-1;
	if (gameix<0 || gameix>=games.num() || !games[gameix]) {
	    message(pl, "5170 No such game.\r\n");
	    return 0;
	}
	int count=0, obscount=0;
	for (int i=0; i<players.num(); i++) {
	    if (players[i] && players[i]->gamenum == gamenum) {
		if (players[i]->ready < 0) obscount++;
		else count++;
	    }
	}
	message(pl, "1170 %d %s %d %d players/observers\r\n", 
		gamenum, games[gameix]->game_started ? "PLAY" : "FORM",
		count, obscount);
    }

    ////////////////////
    // NEWGAME

    else if (!strcasecmp(line, "NEWGAME")) {
        player_unidle(pl);
	if (!pl->hasname) {
	    message(pl, "5182 Use NAME first.\r\n");
	    return 0;
	}
	if (pl->gamenum!=0) {
	    message(pl, "5181 Already in a game.\r\n");
	    return 0;
	}

	int gameix=-1;
	for (int i=games.num()-1; i>=0; i--) if (!games[i]) gameix=i;
	if (gameix<0) {
	    if (games.num() >= MAXGAMESATONCE) {
		message(pl, "5180 No space for another game.\r\n");
		return 0;
	    }
	    gameix = games.num();
	    games.setsize(gameix+1);
	}
	games[gameix] = new game;
	games[gameix]->num = gameix+1;
	randomize_tiles(&games[gameix]->tilequeue);
	pl->gamenum = gameix+1;

	message(pl, "1180 %d New game formed.\r\n", gameix+1);

	broadcast("2014 %s has created game %d\r\n", pl->username,pl->gamenum);
	broadcast("2027 %d New game formed\r\n", gameix+1);
    }

    ////////////////////
    // JOIN

    else if (!strncasecmp(line, "JOIN ", 5)) {
        player_unidle(pl);
	if (!pl->hasname) {
	    message(pl, "5192 Use NAME first.\r\n");
	    return 0;
	}
	char *tmp = line+5;
	int gamenum = atoi(tmp);
	int gameix = gamenum-1;
	if (gameix<0 || gameix>=games.num() || !games[gameix]) {
	    message(pl, "5190 No such game.\r\n");
	    return 0;
	}
	if (pl->gamenum!=0) {
	    message(pl, "5191 Already in a game.\r\n");
	    return 0;
	}

	message(pl, "1190 Ok.\r\n");
	pl->gamenum = gamenum;

	if (games[gameix]->game_started) {
	    /* force observer mode */
	    pl->ready = -1;
	    message(pl, "2020 Game is already in progress.\r\n");
	}

	broadcast("2014 %s has entered game %d\r\n", pl->username, gamenum);
	broadcast("2028 %d gamelist info changed\r\n", gameix+1);
    }

    ////////////////////
    // LEAVE

    else if (!strcasecmp(line, "LEAVE")) {
        player_unidle(pl);
	if (!pl->hasname) {
	    message(pl, "5202 Use NAME first.\r\n");
	    return 0;
	}
	game *g = game_of_player(pl);
	if (!g) {
	    message(pl, "5200 Not in a game.\r\n");
	    return 0;
	}
	if (g->game_started && pl->ready > 0) {
	    message(pl, "5201 You can't leave now.\r\n");
	    return 0;
	}

	message(pl, "1200 Ok.\r\n");

	broadcast("2014 %s has left game %d\r\n", pl->username, pl->gamenum);

	pl->gamenum = 0;
	pl->ready = 0;

	if (!g->game_started) checkstart(g);

	broadcast("2028 %d gamelist info changed\r\n", g->num);
    }

    ////////////////////
    // HELP

    else if (!strcasecmp(line, "HELP")) {
	message(pl,"1990 Help: Commands are:\r\n");
	message(pl,"2999    NAME username      (set your handle)\r\n");
	message(pl,"2999    GLIST              (show game list)\n");
	message(pl,"2999    GINFO gamenum      (get info about game)\r\n");
	message(pl,"2999    NEWGAME            (start a new game)\r\n");
	message(pl,"2999    JOIN gamenum       (enter a game)\r\n");
	message(pl,"2999    LEAVE              (leave your game)\r\n");
	message(pl,"2999    READY              (vote to start game)\r\n");
	message(pl,"2999    OBSERVER           (do not participate)\r\n");
	message(pl,"2999    PLAY x,y,T x,y,T ... (play tiles T at (x,y))\r\n");
	message(pl,"2999    APPROVE            (accept previous play)\r\n");
	message(pl,"2999    WITHDRAW           (take back previous play)\r\n");
	message(pl,"2999    EXCHANGE           (exchange your rack)\r\n");
	message(pl,"2999    FORFEIT            (if you cannot play)\r\n");
	message(pl,"2999    SAY text           (talk to all players)\r\n");
	message(pl,"2999    TELL user text     (talk to one player)\r\n");
	message(pl,"2999    REFRESH            (request full status)\r\n");
	message(pl,"2999    BEEP user          (page a user)\r\n");
	message(pl,"2999    QUIT               (disconnect)\r\n");
	message(pl,"2999 x,y of 0,0 is the top left.\r\n");
	if (!pl->hasname) message(pl,"2999 You must use NAME first.\r\n");
    }

    // Oops.

    else message(pl, "5999 Unknown or malformed command\r\n");
    return 0;
}

static int clientdata(client *c, const char *data, size_t len) {
    int ret = 0;
    assert(c->bufpos+len <= sizeof(c->buffer));
    memcpy(c->buffer+c->bufpos, data, len);
    c->bufpos += len;
    char *s;
    while ((s=(char *)memchr(c->buffer, '\n', c->bufpos))!=NULL) {
	*s = 0;
	if (s!=c->buffer && s[-1]=='\r') s[-1] = 0;
	s++;
	c->bufpos = c->bufpos - (s - c->buffer);
	player *p = players[c->playernum];
	assert(p);
	int r = clientline(c, p, c->buffer);
	if (r) ret = 1;
	memmove(c->buffer, s, c->bufpos);
    }
    return ret;
}

//////////////////////////////////////////////////////////////////////

static void mainloop(int sock) {
    fd_set xset, aset;
    FD_ZERO(&aset);
    FD_SET(sock, &aset);
    int hifd = sock+1;

    bool done=false;
    while (!done) {
	xset = aset;
	if (select(hifd, &xset, NULL, NULL, NULL)<0) {
	    if (errno!=EINTR) {
		printf("select: %s\n", strerror(errno));
	    }
	    continue;
	}
	if (FD_ISSET(sock, &xset)) {
	    struct sockaddr_in sn;
	    socklen_t size=sizeof(sn);
	    int newsock = accept(sock, (struct sockaddr *)&sn, &size);
	    if (size!=sizeof(sn) || sn.sin_family != AF_INET) {
		/* ? */
		close(newsock);
	    }
	    else if (newclient(newsock, &sn)) {
		close(newsock);
	    }
	    else {
		/* This will hopefully time out hung-up PPP links. */
		/* Alas, I suspect it won't be effective. */
		int y=1;
		setsockopt(newsock, SOL_SOCKET, SO_KEEPALIVE, 
			   (char *)&y, sizeof(y));

		FD_SET(newsock, &aset);
		if (newsock >= hifd) hifd = newsock+1;
	    }
	}
	for (int i=0; i<clients.num(); i++) {
	    if (clients[i] && FD_ISSET(clients[i]->fd, &xset)) {
		bool drop = true;
		char buf[4096];
		client *c = clients[i];

		int len = read(c->fd, buf, sizeof(buf));
		if (len<0) {
		    printf("client %s: read: %s\n", c->hostname, 
			   strerror(errno));
		}
		else if (len==0) {
		    printf("client %s: EOF\n", c->hostname);
		}
		else if (clientdata(c, buf, len)) {
		    printf("client %s: QUIT\n", c->hostname);
		}
		else {
		    drop = false;
		}

		if (drop) {
		    // Take it out of the table first so you can broadcast() 
		    // from deadclient().
		    clients[i] = NULL;
		    FD_CLR(c->fd, &aset);
		    deadclient(c);
		    close(c->fd);
		    delete c;
		    for (int j=0; j<games.num(); j++) if (games[j]) {
			checkreset(games[j]);
		    }
		}
	    }
	}
    }
}

static int setup() {
    struct sockaddr_in sn;
    memset(&sn, 0, sizeof(sn));
    sn.sin_family = AF_INET;
    sn.sin_port = htons(SCRAB_PORT);
    sn.sin_addr.s_addr = INADDR_ANY;

    int sock = socket(AF_INET, SOCK_STREAM, 0);
    if (sock<0) {
	printf("socket: %s\n", strerror(errno));
	exit(1);
    }

    int yes=1;
    setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, (char *)&yes, sizeof(yes));

    if (bind(sock, (struct sockaddr *)&sn, sizeof(sn))) {
	printf("bind: %s\n", strerror(errno));
	exit(1);
    }

    if (listen(sock, 128)) {
	printf("listen: %s\n", strerror(errno));
	exit(1);
    }

    return sock;
}

static void disconnect(int sock) {
    broadcast("2015 Server disconnecting\r\n");
    for (int i=0; i<clients.num(); i++) if (clients[i]) {
	close(clients[i]->fd);
    }
    shutdown(sock, 2);
    close(sock);
}

int main(int argc, char *argv[]) {
    int ch;
    while ((ch = getopt(argc, argv, "an"))!=-1) {
        switch (ch) {
	  case 'n': use_dns = 0; break;
	  case 'a': reject_no_dns = 0; break;
	  default:
	    printf("Invalid option -%c\n", ch);
	    printf("Usage: scrabserve [-an]\n");
	    printf("   -a:  allow connections with no reverse DNS\n");
	    printf("   -n:  don't query DNS at all (implies -a)\n");
	    return 1;
	}
    }

    srandom(time(NULL)^(getpid()<<16)^getuid());

    games.setsize(1);
    games[0] = new game;
    games[0]->num = 1;
    randomize_tiles(&games[0]->tilequeue);

    int sock = setup();
    printf("Listening for connections...\n");
    mainloop(sock);
    disconnect(sock);
    return 0;
}
